"""
Tests for ExchangeCalendarDispatcher.
"""

from unittest import TestCase
import re

import pandas as pd
import pytest

from exchange_calendars import ExchangeCalendar
from exchange_calendars.calendar_utils import ExchangeCalendarDispatcher
from exchange_calendars.exchange_calendar_iepa import IEPAExchangeCalendar
from exchange_calendars.errors import (
    CalendarNameCollision,
    CyclicCalendarAlias,
    InvalidCalendarName,
)


class CalendarDispatcherTestCase(TestCase):
    @classmethod
    def setup_class(cls):
        cls.dispatcher_kwargs = {
            "calendars": {},
            "calendar_factories": {"IEPA": IEPAExchangeCalendar},
            "aliases": {
                "IEPA_ALIAS": "IEPA",
                "IEPA_ALIAS_ALIAS": "IEPA_ALIAS",
            },
        }

    def setup_method(self, method):  # noqa: ARG002
        self.dispatcher = ExchangeCalendarDispatcher(
            # Make copies here so that tests that mutate the dispatcher dicts
            # are isolated from one another.
            **{k: v.copy() for k, v in self.dispatcher_kwargs.items()}
        )

    def teardown_method(self, method):  # noqa: ARG002
        self.dispatcher = None

    @classmethod
    def teardown_class(cls):
        cls.dispatcher_kwargs = None

    def test_follow_alias_chain(self):
        self.assertIs(
            self.dispatcher.get_calendar("IEPA_ALIAS"),
            self.dispatcher.get_calendar("IEPA"),
        )
        self.assertIs(
            self.dispatcher.get_calendar("IEPA_ALIAS_ALIAS"),
            self.dispatcher.get_calendar("IEPA"),
        )

    def test_add_new_aliases(self):
        with self.assertRaises(InvalidCalendarName):
            self.dispatcher.get_calendar("NOT_IEPA")

        self.dispatcher.register_calendar_alias("NOT_IEPA", "IEPA")

        self.assertIs(
            self.dispatcher.get_calendar("NOT_IEPA"),
            self.dispatcher.get_calendar("IEPA"),
        )

        self.dispatcher.register_calendar_alias(
            "IEPA_ALIAS_ALIAS_ALIAS", "IEPA_ALIAS_ALIAS"
        )
        self.assertIs(
            self.dispatcher.get_calendar("IEPA_ALIAS_ALIAS_ALIAS"),
            self.dispatcher.get_calendar("IEPA"),
        )

    def test_remove_aliases(self):
        self.dispatcher.deregister_calendar("IEPA_ALIAS_ALIAS")
        with self.assertRaises(InvalidCalendarName):
            self.dispatcher.get_calendar("IEPA_ALIAS_ALIAS")

    def test_reject_alias_that_already_exists(self):
        with self.assertRaises(CalendarNameCollision):
            self.dispatcher.register_calendar_alias("IEPA", "NOT_IEPA")

        with self.assertRaises(CalendarNameCollision):
            self.dispatcher.register_calendar_alias("IEPA_ALIAS", "NOT_IEPA")

    def test_allow_alias_override_with_force(self):
        self.dispatcher.register_calendar_alias("IEPA", "NOT_IEPA", force=True)
        with self.assertRaises(InvalidCalendarName):
            self.dispatcher.get_calendar("IEPA")

    def test_reject_cyclic_aliases(self):
        add_alias = self.dispatcher.register_calendar_alias

        add_alias("A", "B")
        add_alias("B", "C")

        with self.assertRaises(CyclicCalendarAlias) as e:
            add_alias("C", "A")

        expected = "Cycle in calendar aliases: ['C' -> 'A' -> 'B' -> 'C']"
        self.assertEqual(str(e.exception), expected)

    def test_get_calendar_names(self):
        self.assertEqual(
            sorted(self.dispatcher.get_calendar_names()),
            ["IEPA", "IEPA_ALIAS", "IEPA_ALIAS_ALIAS"],
        )
        self.assertEqual(
            self.dispatcher.get_calendar_names(include_aliases=False),
            ["IEPA"],
        )

    def test_aliases_to_names(self):
        self.assertDictEqual(
            self.dispatcher.aliases_to_names(),
            {
                "IEPA_ALIAS": "IEPA",
                "IEPA_ALIAS_ALIAS": "IEPA",
            },
        )

    def test_names_to_aliases(self):
        self.assertDictEqual(
            self.dispatcher.names_to_aliases(),
            {"IEPA": ["IEPA_ALIAS", "IEPA_ALIAS_ALIAS"]},
        )

    def test_get_calendar(self):
        cal = self.dispatcher.get_calendar("IEPA")
        self.assertIsInstance(cal, ExchangeCalendar)

    def test_get_calendar_kwargs(self):
        start = pd.Timestamp("2020-01-02")
        end = pd.Timestamp("2020-01-31")
        cal = self.dispatcher.get_calendar("IEPA", start=start, end=end)
        self.assertEqual(cal.first_session, start)
        self.assertEqual(cal.last_session, end)

        self.dispatcher.register_calendar("iepa_instance", cal)
        error_msg = (
            f"Receieved calendar arguments although iepa_instance is registered"
            f" as a specific instance of class {cal.__class__}, not as a"
            f" calendar factory."
        )
        with pytest.raises(ValueError, match=re.escape(error_msg)):
            # Can only pass kwargs to registered factories (not calendar instances)
            self.dispatcher.get_calendar("iepa_instance", start=start)

        with pytest.raises(ValueError, match=re.escape(error_msg)):
            # Can only pass kwargs to registered factories (not calendar instances)
            self.dispatcher.get_calendar("iepa_instance", side="right")

    def test_get_calendar_cache(self):
        start = pd.Timestamp("2020-01-02")
        end = pd.Timestamp("2020-01-31")
        cal = self.dispatcher.get_calendar("IEPA", start=start, end=end, side="right")
        cal2 = self.dispatcher.get_calendar("IEPA", start=start, end=end, side="right")
        self.assertIs(cal, cal2)
        start += pd.DateOffset(days=1)
        cal3 = self.dispatcher.get_calendar("IEPA", start=start, end=end, side="right")
        self.assertIsNot(cal, cal3)
        cal4 = self.dispatcher.get_calendar("IEPA", start=start, end=end, side="right")
        self.assertIs(cal3, cal4)
        cal5 = self.dispatcher.get_calendar("IEPA", start=start, end=end, side="left")
        self.assertIsNot(cal4, cal5)



import pytest

from exchange_calendars.exchange_calendar_xstu import XSTUExchangeCalendar
from .test_exchange_calendar import ExchangeCalendarTestBase


class TestXSTUCalendar(ExchangeCalendarTestBase):
    @pytest.fixture(scope="class")
    def calendar_cls(self):
        yield XSTUExchangeCalendar

    @pytest.fixture
    def max_session_hours(self):
        # The XSTU is open from 7:30 am to 10:00 pm.
        yield 14.5

    @pytest.fixture
    def regular_holidays_sample(self):
        yield [
            # 2025
            "2025-01-01",  # New Year
            "2025-04-18",  # Good Friday
            "2025-04-21",  # Easter Monday
            "2025-05-01",  # Labor Day
            "2025-12-24",  # Christmas Eve
            "2025-12-25",  # Christmas
            "2025-12-26",  # Boxing Day
            "2025-12-31",  # New Year's Eve
            # 2026
            "2026-01-01",  # New Year
            "2026-04-03",  # Good Friday
            "2026-04-06",  # Easter Monday
            "2026-05-01",  # Labor Day
            "2026-12-24",  # Christmas Eve
            "2026-12-25",  # Christmas
            # Boxing Day 2026 is Sat, so not a trading day anyway
            "2026-12-31",  # New Year's Eve
        ]

    @pytest.fixture
    def adhoc_holidays_sample(self):
        yield []

    @pytest.fixture
    def non_holidays_sample(self):
        yield [
            # Check Whit Monday and German Unity Day are NOT holidays
            "2025-06-09",  # Whit Monday
            "2025-10-03",  # German Unity Day
            "2026-05-25",  # Whit Monday
        ]

    @pytest.fixture
    def early_closes_sample(self):
        yield []

    @pytest.fixture
    def early_closes_sample_time(self):
        yield None



import pytest

from exchange_calendars.exchange_calendar_xlit import XLITExchangeCalendar

from .test_exchange_calendar import ExchangeCalendarTestBase


class TestXLITExchangeCalendar(ExchangeCalendarTestBase):
    @pytest.fixture(scope="class")
    def calendar_cls(self):
        return XLITExchangeCalendar

    @pytest.fixture
    def max_session_hours(self):
        # The XLIT is open from 10:00 AM to 4:00 PM on its longest trading day.
        yield 6

    @pytest.fixture
    def regular_holidays_sample(self):
        yield [
            "2024-01-01",  # New Year's Day
            "2024-02-16",  # Restoration of the State
            "2024-03-11",  # Restoration of Independence
            "2024-03-29",  # Good Friday
            "2024-04-01",  # Easter Monday
            "2024-05-01",  # Labour Day
            "2024-05-09",  # Ascension Day
            "2024-06-24",  # St. John's Day
            "2024-07-06",  # Statehood Day
            "2024-08-15",  # Assumption Day
            "2024-11-01",  # All Saints' Day
            "2024-12-24",  # Christmas Eve
            "2024-12-26",  # Boxing Day
            "2024-12-31",  # New Year's Eve
        ]

    @pytest.fixture
    def non_holidays_sample(self):
        yield [
            "2024-01-02",  # Not a holiday
            "2024-03-12",  # Not a holiday
            "2024-04-02",  # Not a holiday
            "2024-05-02",  # Not a holiday
            "2024-06-25",  # Not a holiday
        ]



import itertools

import pandas as pd

from exchange_calendars import get_calendar
from exchange_calendars.calendar_helpers import UTC
from exchange_calendars.calendar_utils import (
    ExchangeCalendarDispatcher,
    _default_calendar_aliases,
    _default_calendar_factories,
)


def construct_all_calendars():
    dispatcher = ExchangeCalendarDispatcher(
        calendars={},
        calendar_factories=_default_calendar_factories,
        aliases=_default_calendar_aliases,
    )

    calendar_names = itertools.chain(
        _default_calendar_aliases, _default_calendar_factories
    )

    for name in calendar_names:
        assert dispatcher.get_calendar(name) is not None
        dispatcher.deregister_calendar(name)


def is_open_on_minute_bench(cal, timestamps):
    for timestamp in timestamps:
        cal.is_open_on_minute(timestamp, _parse=False)


def test_calendar_construction(benchmark):
    benchmark(construct_all_calendars)


def test_is_open_on_minute(benchmark):
    xhkg = get_calendar("XHKG")
    timestamps = [
        pd.Timestamp("2019-10-11 01:20:00", tz=UTC),  # pre open
        pd.Timestamp("2019-10-11 01:30:00", tz=UTC),  # open
        pd.Timestamp("2019-10-11 01:31:00", tz=UTC),  # first minute
        pd.Timestamp("2019-10-11 04:31:00", tz=UTC),  # in break
        pd.Timestamp("2019-10-11 08:00:00", tz=UTC),  # close
        pd.Timestamp("2019-10-11 08:01:00", tz=UTC),  # post close
    ]
    benchmark(is_open_on_minute_bench, xhkg, timestamps)



import pandas as pd
import pytest

from exchange_calendars.exchange_calendar_xasx import XASXExchangeCalendar

from .test_exchange_calendar import ExchangeCalendarTestBase


class TestXASXCalendar(ExchangeCalendarTestBase):
    @pytest.fixture(scope="class")
    def calendar_cls(self):
        yield XASXExchangeCalendar

    @pytest.fixture
    def max_session_hours(self):
        yield 6

    @pytest.fixture
    def regular_holidays_sample(self):
        yield [
            # 2018
            "2018-01-01",  # New Year's Day
            "2018-01-26",  # Australia Day
            "2018-03-30",  # Good Friday
            "2018-04-02",  # Easter Monday
            "2018-04-25",  # Anzac Day
            "2018-06-11",  # Queen's Birthday
            "2018-12-25",  # Christmas
            "2018-12-26",  # Boxing Day
            #
            # Holidays made up when fall on weekend.
            # Anzac Day is observed on the following Monday only when falling
            # on a Sunday. In years where Anzac Day falls on a Saturday, there
            # is no make-up.
            "2017-01-02",  # New Year's Day on a Sunday, observed on Monday.
            "2014-01-27",  # Australia Day on a Sunday, observed on Monday (from 2010).
            "2010-04-26",  # Anzac Day on a Sunday, observed on Monday.
            # Christmas/Boxing Day are special cases, whereby if Christmas is a
            # Saturday and Boxing Day is a Sunday, the next Monday and Tuesday will
            # be holidays. If Christmas is a Sunday and Boxing Day is a Monday then
            # Monday and Tuesday will still both be holidays.
            # Christmas on a Sunday, Boxing Day on Monday.
            "2016-12-26",
            "2016-12-27",
            # Christmas on a Saturday, Boxing Day on Sunday.
            "2010-12-27",
            "2010-12-28",
        ]

    @pytest.fixture
    def adhoc_holidays_sample(self):
        # National day of mourning for the queen
        yield ["2022-09-22"]

    @pytest.fixture
    def non_holidays_sample(self):
        # Anzac Day on a Saturday, does not have a make-up (prior to 2010).
        yield ["2015-04-27", "2004-04-26"]

    @pytest.fixture
    def early_closes_sample(self):
        yield [
            # In 2018, the last trading days before Christmas and New Year's
            # are on Mondays, so they should be early closes.
            "2018-12-24",
            "2018-12-31",
            # In 2017, Christmas and New Year's fell on Mondays, so the last
            # trading days before them were Fridays, which should be early closes.
            "2017-12-22",
            "2017-12-29",
            # In 2016, Christmas and New Year's fell on Sundays, so the last
            # trading days before them were Fridays, which should be early closes.
            "2016-12-23",
            "2016-12-30",
            # In 2023, the early closes were reinstated after 2022's cancellation.
            "2023-12-22",
            "2023-12-29",
        ]

    @pytest.fixture
    def early_closes_sample_time(self):
        yield pd.Timedelta(hours=14, minutes=10)

    @pytest.fixture
    def non_early_closes_sample(self):
        # In 2009 the early close rules should not be in effect yet.
        # In 2022 we had an exception to them.
        yield ["2009-12-24", "2009-12-31", "2022-12-23", "2022-12-30"]

    @pytest.fixture
    def non_early_closes_sample_time(self):
        yield pd.Timedelta(16, "h")



import pytest

from exchange_calendars.exchange_calendar_xbkk import XBKKExchangeCalendar
from .test_exchange_calendar import ExchangeCalendarTestBase


class TestXBKKCalendar(ExchangeCalendarTestBase):
    @pytest.fixture(scope="class")
    def calendar_cls(self):
        yield XBKKExchangeCalendar

    @pytest.fixture
    def max_session_hours(self):
        yield 6.5

    @pytest.fixture
    def regular_holidays_sample(self):
        yield [
            "2019-01-01",  # New Year's Day
            "2019-02-19",  # Makha Bucha
            "2018-04-06",  # Chakri Memorial Day
            "2016-04-13",  # Songkran Festival
            "2016-04-14",  # Songkran Festival
            "2016-04-15",  # Songkran Festival
            "2019-05-01",  # Labour Day
            "2016-05-05",  # Coronation Day
            "2019-05-20",  # Vesak
            "2019-06-03",  # Queen's Birthday
            "2019-07-16",  # Asanha Bucha
            "2017-07-28",  # King's Birthday
            "2019-08-12",  # Queen Mother's Birthday
            "2017-10-13",  # Passing of King Bhumibol
            "2019-10-23",  # Chulalongkorn Day
            "2019-12-05",  # King Bhumibol's Birthday
            "2019-12-10",  # Thailand Constitution Day
            "2019-12-31",  # New Year's Eve
            #
            # Holidays falling on a weekend and subsequently made up.
            # New Year's Eve on a Saturday and New Year's Day on a Sunday.
            "2017-01-02",  # following Monday a holiday...
            "2017-01-03",  # and following Tuesday.
            "2011-01-03",  # New Year's Day on a Saturday.
            "2019-04-08",  # Chakri Memorial Day (April 6th) on a Saturday.
            # Songkran Festival Days 2 and 3 (Saturday April 14th and Sunday 15th)...
            "2018-04-16",  # although only one day made up, on Monday.
            "2016-05-02",  # Labour Day (May 1st) on a Sunday.
            "2019-05-06",  # Coronation Day (May 5th) on a Sunday.
            "2014-05-05",  # Coronation Day (May 4th) on a Sunday.
            "2019-07-29",  # King's Birthday (July 28th) on a Sunday.
            "2018-08-13",  # Queen Mother's Birthday (August 12th) on a Sunday.
            "2019-10-14",  # The Passing of King Bhumibol (October 13th) on a Sunday.
            "2016-10-24",  # Chulalongkorn Day (October 23rd) on a Sunday.
            "2015-12-07",  # King Bhumibol's Birthday (December 5th) on a Saturday.
            "2017-12-11",  # Thailand Constitution Day (December 10th) on a Sunday.
        ]

    @pytest.fixture
    def adhoc_holidays_sample(self):
        yield [
            # "bridge days" where a Monday or Friday is made into a holiday
            # to fill gap between weekend and a holiday on Tuesday or Thursday.
            "2015-01-02",  # New Year's is a Thusday, so Friday is bridge day.
            "2013-12-30",  # New Year's Eve is a Tuesday, so Monday is bridge day.
            "2016-07-18",  # Asanha Bucha is a Tuesday, so Monday is a bridge day.
            "2010-08-13",  # Queen's B'day a Thursday, so Friday is a bridge day.
            "2016-05-06",  # Coronation Day a Thursday, so Friday is a bridge day.
            "2011-05-16",  # Vesak Day a Tuesday, so the Monday is a bridge day.
        ]

    @pytest.fixture
    def non_holidays_sample(self):
        yield [
            # When Songkran Festival Days 2 and 3 (April 14th and 15th) fall on
            # a Saturday and Sunday, only the next Monday is a holiday...
            "2018-04-17",  # Tuesday 17th is not a holiday.
            # When the last day of Songkran Festival (April 15th) falls on a
            # Saturday...
            "2017-04-17",  # not made up - Monday is not a holdiay.
        ]



import pytest

from exchange_calendars.exchange_calendar_xeur import XEURExchangeCalendar
from .test_exchange_calendar import ExchangeCalendarTestBase


class TestXEURCalendar(ExchangeCalendarTestBase):
    @pytest.fixture(scope="class")
    def calendar_cls(self):
        yield XEURExchangeCalendar

    @pytest.fixture
    def max_session_hours(self):
        # Open from 8:00 am to 10:00 pm.
        yield 14

    @pytest.fixture
    def regular_holidays_sample(self):
        yield [
            # 2026
            "2026-01-01",  # New Year's Day
            "2026-04-03",  # Good Friday
            "2026-04-06",  # Easter Monday
            "2026-05-01",  # Labour Day
            "2026-12-24",  # Christmas Eve
            "2026-12-25",  # Christmas
            # "2026-12-26",  Boxing Day is Saturday
            "2026-12-31",  # New Year's Eve
            # 2027
            "2027-01-01",  # New Year's Day
            "2027-03-26",  # Good Friday
            "2027-03-29",  # Easter Monday
            # "2027-05-01",  Labour day is Saturday
            "2027-12-24",
            # "2027-12-25",  Christmas is Saturday
            # "2027-12-26",  Boxing Day is Sunday
            "2027-12-31",
            # 2028
            # "2028-01-01",  New Year's Day is Saturday
            "2028-04-14",  # Good Friday
            "2028-04-17",  # Easter Monday
            "2028-05-01",  # Labour Day
            # "2028-12-24",  Christmas Eve is Sunday
            "2028-12-25",  # Christmas
            "2028-12-26",  # Boxing Day
            # "2028-12-31",  New Year's Eve is Sunday
        ]

    @pytest.fixture
    def non_holidays_sample(self):
        yield [
            # Whit Monday
            "2026-05-25",
            "2027-05-17",
            # German Unity Day
            "2028-10-03",  # Tuesday
        ]



from datetime import time
from zoneinfo import ZoneInfo
from .pandas_extensions.offsets import OrthodoxEaster
from pandas.tseries.holiday import (
    Holiday,
    EasterMonday,
    GoodFriday,
)
from pandas.tseries.offsets import Day
from .common_holidays import (
    new_years_day,
    epiphany,
    orthodox_good_friday,
    orthodox_easter_monday,
    orthodox_easter_tuesday,
    european_labour_day,
    assumption_day,
    christmas_eve,
    christmas,
    boxing_day,
)
from .exchange_calendar import HolidayCalendar, ExchangeCalendar

NewYearsDay = new_years_day()
Epiphany = epiphany()
GreekIndependenceDay = Holiday("Greek Independence Day", month=3, day=25)
GreenMonday = Holiday(
    "Green Monday", month=1, day=1, offset=[OrthodoxEaster(), -Day(48)]
)
CyprusNationalDay = Holiday("Cyprus National Day", month=4, day=1)
OrthodoxGoodFriday = orthodox_good_friday()
OrthodoxEasterMonday = orthodox_easter_monday()
OrthodoxEasterTuesday = orthodox_easter_tuesday()
LabourDay = european_labour_day()
HolySpiritDay = Holiday(
    "Holy Spirit Day", month=1, day=1, offset=[OrthodoxEaster(), Day(50)]
)
AssumptionDay = assumption_day()
CyprusIndependenceDay = Holiday("Cyprus Independence Day", month=10, day=1)
OkhiDay = Holiday("Okhi Day", month=10, day=28)
ChristmasEve = christmas_eve()
Christmas = christmas()
BoxingDay = boxing_day()


class XCYSExchangeCalendar(ExchangeCalendar):
    """
    Calendar for the Cyprus Stock Exchange.
    https://www.cse.com.cy/en-GB/regulated-market/market-indices/other-information/Public-Holiday/

    Open Time: 10:30 AM EET (Eastern European Time)
    Close Time: 5:00 PM EET (Eastern European Time)

    Regularly-Observed Holidays:
      - New Year's Day
      - Epiphany (Jan 6)
      - Green Monday (Orthodox)
      - Cyprus National Day (Apr 1)
      - Greek Independence Day (Mar 25)
      - Good Friday (Catholic)
      - Easter Monday (Catholic)
      - Good Friday (Orthodox)
      - Easter Monday (Orthodox)
      - Easter Tuesday (Orthodox)
      - Labour Day (May 1)
      - Holy Spirit Day (Orthodox)
      - Assumption Day (Aug 15)
      - Cyprus Independence Day (Oct 1)
      - Okhi Day (Oct 28)
      - Christmas Eve (Dec 24)
      - Christmas Day (Dec 25)
      - Boxing Day (Dec 26)
    """

    name = "XCYS"
    tz = ZoneInfo("Asia/Nicosia")
    open_times = ((None, time(10, 30)),)
    close_times = ((None, time(17, 00)),)

    @property
    def regular_holidays(self):
        return HolidayCalendar(
            [
                NewYearsDay,
                Epiphany,
                GreenMonday,
                CyprusNationalDay,
                GreekIndependenceDay,
                GoodFriday,
                EasterMonday,
                OrthodoxGoodFriday,
                OrthodoxEasterMonday,
                OrthodoxEasterTuesday,
                LabourDay,
                HolySpiritDay,
                AssumptionDay,
                CyprusIndependenceDay,
                OkhiDay,
                ChristmasEve,
                Christmas,
                BoxingDay,
            ]
        )

    @property
    def adhoc_holidays(self):
        return []



from datetime import time
from zoneinfo import ZoneInfo

import pandas as pd
from pandas.tseries.holiday import (
    Holiday,
    next_monday,
    EasterMonday,
    GoodFriday,
)

from .common_holidays import (
    ascension_day,
    boxing_day,
    christmas_eve,
    european_labour_day,
    new_years_day,
    new_years_eve,
)
from .exchange_calendar import HolidayCalendar, ExchangeCalendar

NewYearsDay = new_years_day()
LabourDay = european_labour_day()
AscensionDay = ascension_day()
ChristmasEve = christmas_eve()
BoxingDay = boxing_day()
NewYearsEve = new_years_eve()

RestorationOfIndependence = Holiday(
    "Restoration of Independence", month=5, day=4, observance=next_monday
)

MidsummerEve = Holiday("Midsummer's Eve", month=6, day=23)

MidsummerDay = Holiday("Midsummer's Day", month=6, day=24)

ProclamationDay = Holiday("Proclamation Day", month=11, day=18, observance=next_monday)


class XRISExchangeCalendar(ExchangeCalendar):
    """
    Calendar for the Nasdaq Riga Stock Exchange (Latvia).
    https://nasdaqbaltic.com/statistics/en/calendar?holidays=1

    Open Time: 10:00 AM, EET (Eastern European Time)
    Close Time: 4:00 PM, EET (Eastern European Time)

    Regularly-Observed Holidays:
      - New Year's Day
      - Good Friday
      - Easter Monday
      - Labour Day (May 1)
      - Restoration of Independence (May 4)
      - Ascension Day
      - Midsummer's Eve (Jun 23)
      -	Midsummer's Day (Jun 24)
      - Proclamation Day (Nov 18)
      - Christmas Eve (Dec 24)
      - Boxing Day (Dec 26)
      - New Year's Eve (Dec 31)
    """

    name = "XRIS"
    tz = ZoneInfo("Europe/Riga")
    open_times = ((None, time(10, 0)),)
    close_times = ((None, time(16, 0)),)

    @property
    def regular_holidays(self):
        return HolidayCalendar(
            [
                NewYearsDay,
                GoodFriday,
                EasterMonday,
                LabourDay,
                RestorationOfIndependence,
                AscensionDay,
                MidsummerEve,
                MidsummerDay,
                ProclamationDay,
                ChristmasEve,
                BoxingDay,
                NewYearsEve,
            ]
        )

    @property
    def adhoc_holidays(self):
        return [
            pd.Timestamp(
                "2028-07-10"
            ),  # National Holiday - Latvian Song and Dance Festival (every 5 years)
            pd.Timestamp("2025-11-17"),  # Additional Day off - Latvia's National Day
            pd.Timestamp("2025-05-02"),  # Additional Day off
            pd.Timestamp(
                "2024-12-30"
            ),  # Additional Day off - Compensation by 28 December
            pd.Timestamp(
                "2024-12-23"
            ),  # Additional Day off - Compensation by 14 December
            pd.Timestamp("2023-12-25"),  # Additional Day off
            pd.Timestamp(
                "2023-07-10"
            ),  # National Holiday - Latvian Song and Dance Festival (every 5 years)
            pd.Timestamp(
                "2023-05-05"
            ),  # Additional Day off - Compensation by Saturday 20 May (bridge day)
            pd.Timestamp(
                "2018-05-09"
            ),  # National Holiday - Latvian Song and Dance Festival (every 5 years)
        ]



import pytest
from exchange_calendars.exchange_calendar_xlju import XLJUExchangeCalendar
from .test_exchange_calendar import ExchangeCalendarTestBase


class TestXLJUCalendar(ExchangeCalendarTestBase):
    @pytest.fixture(scope="class")
    def calendar_cls(self):
        yield XLJUExchangeCalendar

    @pytest.fixture
    def max_session_hours(self):
        # The XLJU is open from 9:15 am to 3:15 pm.
        yield 6.0

    @pytest.fixture
    def regular_holidays_sample(self):
        yield [
            "2025-01-01",  # New Year's Day
            "2025-01-02",  # New Year Holiday
            "2025-02-08",  # Prešeren Day
            "2025-04-18",  # Good Friday
            "2025-04-21",  # Easter Monday
            "2025-04-27",  # Resistance Day
            "2025-05-01",  # Labour Day
            "2025-05-02",  # Labour Day (Second Day)
            "2025-06-25",  # Statehood Day
            "2025-08-15",  # Assumption Day
            "2025-10-31",  # Reformation Day
            "2025-11-01",  # All Saints' Day
            "2025-12-24",  # Christmas Day
            "2025-12-25",  # Christmas Day
            "2025-12-26",  # Independence and Unity Day
            "2025-12-31",  # New Year's Eve
        ]

    @pytest.fixture
    def non_holidays_sample(self):
        yield [
            "2025-01-03",
            "2025-04-22",
            "2025-06-26",
            "2025-08-14",
            "2025-12-30",
        ]



import io
import sys

import pandas as pd

# ruff: noqa: T201

months = [
    "January",
    "February",
    "March",
    "April",
    "May",
    "June",
    "July",
    "August",
    "September",
    "October",
    "November",
    "December",
]


def error(msg):
    print(msg, file=sys.stderr)
    sys.exit(-1)


def _render_month(calendar, year, month, print_year):
    out = io.StringIO()

    start = f"{year}-{month}"
    end = f"{year + 1}-{1}" if month == 12 else f"{year}-{month + 1}"

    days = pd.date_range(start, end, inclusive="left")

    title = months[month - 1]
    if print_year:
        title += f" {year}"

    print(f"{title:^28}".rstrip(), file=out)
    print(" Su  Mo  Tu  We  Th  Fr  Sa", file=out)
    print(
        " " * (4 * ((days[0].weekday() + 1) % 7)),
        end="",
        file=out,
    )

    for d in days:
        if d.weekday() == 6:
            print("", file=out)  # noqa: FURB105

        if calendar.is_session(d):
            a = b = " "
        else:
            a = "["
            b = "]"

        print(
            f"{a}{d.day:>2}{b}",
            end="",
            file=out,
        )

    print("", file=out)  # noqa: FURB105
    return out.getvalue()


def _concat_lines(strings, width):
    as_lines = [string.splitlines() for string in strings]
    max_lines = max(len(lines) for lines in as_lines)
    for lines in as_lines:
        missing_lines = max_lines - len(lines)
        if missing_lines:
            lines.extend([" " * width] * missing_lines)

    rows = []
    for row_parts in zip(*as_lines, strict=False):
        row_parts = list(row_parts)  # noqa: PLW2901
        for n, row_part in enumerate(row_parts):
            missing_space = width - len(row_part)
            if missing_space:
                row_parts[n] = row_part + " " * missing_space

        rows.append("   ".join(row_parts))

    return "\n".join(row.rstrip() for row in rows)


def _int_arg(v, name):
    try:
        return int(v)
    except ValueError:
        error(f"{name} must be an integer, got: {v}")


def parse_args(argv):
    usage = f"usage: {argv[0]} CALENDAR [[[DAY] MONTH] YEAR]"

    if len(argv) == 1 or "--help" in argv or "-h" in argv:
        error(usage)

    if len(argv) > 1:
        from exchange_calendars import get_calendar  # noqa: PLC0415

        try:
            calendar = get_calendar(argv[1])
        except Exception as e:  # noqa: BLE001
            error(str(e))

    if len(argv) == 2:
        import datetime  # noqa: PLC0415

        now = datetime.datetime.now()
        year = now.year
        month = now.month
    elif len(argv) == 3:
        year = _int_arg(argv[2], "YEAR")
        month = None
    elif len(argv) == 4:
        month = _int_arg(argv[2], "MONTH")
        year = _int_arg(argv[3], "YEAR")
    else:
        error(usage)

    return calendar, year, month


def main(argv=None):
    """Print a unix-cal like calendar but indicate which days are trading
    sessions.
    """
    if argv is None:
        argv = sys.argv
    calendar, year, month = parse_args(argv)

    if month is not None:
        print(_render_month(calendar, year, month, print_year=True))
    else:
        month_strings = [
            [
                _render_month(
                    calendar,
                    year,
                    row * 3 + column + 1,
                    print_year=False,
                )
                for column in range(3)
            ]
            for row in range(4)
        ]
        print(f"{year:^88}\n".rstrip())
        print("\n\n".join(_concat_lines(cs, 28) for cs in month_strings))


if __name__ == "__main__":
    main()



import pytest

from exchange_calendars.exchange_calendar_xeee import XEEEExchangeCalendar
from .test_exchange_calendar import ExchangeCalendarTestBase


class TestXEEECalendar(ExchangeCalendarTestBase):
    @pytest.fixture(scope="class")
    def calendar_cls(self):
        yield XEEEExchangeCalendar

    @pytest.fixture
    def max_session_hours(self):
        # The XEEE is open from 8:00 am to 6:00 pm.
        yield 10

    @pytest.fixture
    def regular_holidays_sample(self):
        yield [
            # 2018
            "2018-01-01",  # New Year's Day
            "2018-03-30",  # Good Friday
            "2018-04-02",  # Easter Monday
            "2018-05-01",  # Labor Day
            "2018-12-24",  # Christmas Eve
            "2018-12-25",  # Christmas Day
            "2018-12-26",  # Boxing Day
            "2018-12-31",  # New Year's Eve
        ]

    @pytest.fixture
    def non_holidays_sample(self):
        yield [
            # Holidays that fall on a weekend and are not made up. Ensure surrounding
            # days are not holidays.
            # In 2017 New Year's Day fell on a Sunday, so the Friday before and
            # the Monday after should both be open.
            "2016-12-30",
            "2017-01-02",
            # In 2010 Labour Day fell on a Saturday, so the market should be
            # open on both the prior Friday and the following Monday.
            "2010-04-30",
            "2010-05-03",
            # Christmas also fell on a Saturday, meaning Boxing Day fell on a
            # Sunday. The market should be closed on Christmas Eve (Friday) but
            # still be open on both the prior Thursday and the next Monday.
            "2010-12-23",
            "2010-12-27",
        ]



#
# Copyright 2018 Quantopian, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from datetime import time
from zoneinfo import ZoneInfo

import pandas as pd
import functools

from pandas.tseries.offsets import CustomBusinessDay

from exchange_calendars.pandas_extensions.holiday import AbstractHolidayCalendar
from .tase_holidays import (
    FastDay,
    IndependenceDay,
    MemorialDay,
    NewYear,
    NewYear2,
    NewYearsEve,
    Passover,
    Passover2,
    PassoverInterimDay1,
    PassoverInterimDay1Before2026,
    PassoverInterimDay2,
    PassoverInterimDay2Before2026,
    PassoverInterimDay3,
    PassoverInterimDay3Before2026,
    PassoverInterimDay4,
    PassoverInterimDay4Before2026,
    SukkothInterimDay1,
    SukkothInterimDay1Before2026,
    SukkothInterimDay2,
    SukkothInterimDay2Before2026,
    SukkothInterimDay3,
    SukkothInterimDay3Before2026,
    SukkothInterimDay4,
    SukkothInterimDay4Before2026,
    SukkothInterimDay5,
    SukkothInterimDay5Before2026,
    Passover2Eve,
    PassoverEve,
    Pentecost,
    PentecostEve,
    Purim,
    SimchatTorah,
    SimchatTorahEve,
    Sukkoth,
    SukkothEve,
    YomKippur,
    YomKippurEve,
    YomKippurEveObserved,
)
from .exchange_calendar import HolidayCalendar, ExchangeCalendar
from .pandas_extensions.offsets import MultipleWeekmaskCustomBusinessDay

# All holidays are defined as ad-hoc holidays for each year since there is
# currently no support for Hebrew calendar holiday rules in pandas.

HOLIDAY_EARLY_CLOSE = time(14, 15)
SUNDAY_CLOSE = time(15, 40)
FRIDAY_CLOSE = time(13, 50)


class SundayUntil2026(HolidayCalendar):
    def __init__(self):
        super().__init__(rules=[])

    def holidays(self, start=None, end=None, return_name=False):  # noqa: ARG002
        limit = pd.Timestamp("2026-01-05")
        if start is None:
            start = pd.Timestamp(AbstractHolidayCalendar.start_date)
        if end is None:
            end = pd.Timestamp(AbstractHolidayCalendar.end_date)
        effective_end = min(end, limit)
        if start > effective_end:
            return pd.DatetimeIndex([])
        return pd.date_range(start, effective_end, freq="W-SUN")


class FridayFrom2026(HolidayCalendar):
    def __init__(self):
        super().__init__(rules=[])

    def holidays(self, start=None, end=None, return_name=False):  # noqa: ARG002
        limit = pd.Timestamp("2026-01-05")
        if start is None:
            start = pd.Timestamp(AbstractHolidayCalendar.start_date)
        if end is None:
            end = pd.Timestamp(AbstractHolidayCalendar.end_date)
        effective_start = max(start, limit)
        if effective_start > end:
            return pd.DatetimeIndex([])
        return pd.date_range(effective_start, end, freq="W-FRI")


class XTAEExchangeCalendar(ExchangeCalendar):
    """
    Exchange calendar for TASE Stock Exchange.

    Open/close times are continuous trading times valid Mon-Thu for Shares
    Group A. Trading schedule differs on Sundays.

    Open Time: 9:59 AM, Asia/Tel_Aviv (randomly between 9:59 and 10:00).
    Close Time: 5:15 PM, Asia/Tel_Aviv (randomly between 5:14 and 5:15).
    Friday close time: 1:50pm, Asia/Tel_Aviv

    Regularly-Observed Holidays (not necessarily in order):
    - Purim
    - Passover Eve
    - Passover
    - Passover II Eve
    - Passover II
    - Memorial Day
    - Independence Day
    - Pentecost Eve
    - Pentecost
    - Tisha B'Av
    - New Year's Eve
    - New Year
    - New Year II
    - Yom Kippur Eve
    - Yom Kippur
    - Sukkoth Eve
    - Sukkoth
    - Simchat Torah Eve
    - Simchat Torah

    Note these dates are only checked against 2019-2026.

    https://www.tase.co.il/en/content/knowledge_center/trading_vacation_schedule

    Daylight Saving Time in Israel comes into effect on the Friday before the
    last Sunday in March, and lasts until the last Sunday in October. During the
    Daylight Saving time period the clock will be UTC+3, and UTC+2 for the rest
    of the year.

    Transition to Monday-Friday Trading
    https://www.tase.co.il/en/content/about/tradingdays_change
    """

    name = "XTAE"

    tz = ZoneInfo("Asia/Tel_Aviv")

    open_times = ((None, time(9, 59)),)

    close_times = ((None, time(17, 15)),)

    @property
    def regular_holidays(self):
        return HolidayCalendar(
            [
                Purim,
                PassoverEve,
                Passover,
                Passover2Eve,
                Passover2,
                MemorialDay,
                IndependenceDay,
                PentecostEve,
                Pentecost,
                FastDay,
                NewYearsEve,
                NewYear,
                NewYear2,
                YomKippurEve,
                YomKippurEveObserved,
                YomKippur,
                SukkothEve,
                Sukkoth,
                SimchatTorahEve,
                SimchatTorah,
            ]
        )

    @property
    def adhoc_holidays(self):
        return [
            # Election days:
            # - 2019
            pd.Timestamp("2019-04-09"),
            pd.Timestamp("2019-09-17"),
            # - 2020
            pd.Timestamp("2020-03-02"),
            # - 2021
            pd.Timestamp("2021-03-23"),
            # - 2022
            pd.Timestamp("2022-11-01"),
        ]

    @property
    def special_closes(self):
        return [
            (
                HOLIDAY_EARLY_CLOSE,
                HolidayCalendar(
                    [
                        PassoverInterimDay1,
                        PassoverInterimDay2,
                        PassoverInterimDay3,
                        PassoverInterimDay4,
                        PassoverInterimDay1Before2026,
                        PassoverInterimDay2Before2026,
                        PassoverInterimDay3Before2026,
                        PassoverInterimDay4Before2026,
                        SukkothInterimDay1,
                        SukkothInterimDay2,
                        SukkothInterimDay3,
                        SukkothInterimDay4,
                        SukkothInterimDay5,
                        SukkothInterimDay1Before2026,
                        SukkothInterimDay2Before2026,
                        SukkothInterimDay3Before2026,
                        SukkothInterimDay4Before2026,
                        SukkothInterimDay5Before2026,
                    ]
                ),
            ),
            (SUNDAY_CLOSE, SundayUntil2026()),
            (FRIDAY_CLOSE, FridayFrom2026()),
        ]

    @property
    def special_weekmasks(self):
        """
        Returns
        -------
        list: List of (date, date, str) tuples that represent special
         weekmasks that applies between dates.
        """
        return [
            (None, pd.Timestamp("2026-01-04"), "1111001"),
        ]

    @functools.cached_property
    def day(self):
        if self.special_weekmasks:
            return MultipleWeekmaskCustomBusinessDay(
                holidays=self.adhoc_holidays,
                calendar=self.regular_holidays,
                weekmask=self.weekmask,
                weekmasks=self.special_weekmasks,
            )
        return CustomBusinessDay(
            holidays=self.adhoc_holidays,
            calendar=self.regular_holidays,
            weekmask=self.weekmask,
        )



import pytest
import pandas as pd

from exchange_calendars.exchange_calendar_xtks import XTKSExchangeCalendar
from exchange_calendars.exchange_calendar import SUNDAY, WEDNESDAY
from exchange_calendars.xtks_holidays import (
    AutumnalEquinoxes,
    ChildrensDay,
    CitizensHolidaySilverWeek,
    ConstitutionMemorialDay,
    EmperorAkihitoBirthday,
    GreeneryDay2007Onwards,
    RespectForTheAgedDay2003Onwards,
)
from .test_exchange_calendar import ExchangeCalendarTestBase
from .test_utils import T

# NOTE: A couple of dedicated tests, and sanity tests in fixtures, test
# holidays imported from xtks_holidays. These tests would probably be
# better placed on a spearate test module, or at least to a separate
# class on this module.


class TestXTKSCalendar(ExchangeCalendarTestBase):
    @pytest.fixture(scope="class")
    def calendar_cls(self):
        yield XTKSExchangeCalendar

    @pytest.fixture
    def max_session_hours(self):
        yield 6.5

    @pytest.fixture
    def start_bound(self):
        yield T("1997-01-01")

    @pytest.fixture  # Calendar-specific fixture
    def silver_week_holidays(self):
        """List of pd.Timestamp representing Citizens Day and
        Respect for the Aged Day holidays of silver week.
        """

        def day_before(dt):
            return dt - pd.Timedelta(days=1)

        # Make sure that every Tuesday between Respect for the Aged Day and
        # the Autumnal Equinox is a citizen's holiday
        citizens_holidays = []
        for equinox in AutumnalEquinoxes:
            # It is unusual for September to get this extra holiday
            # so the presence of a "silver week" was not widely noted before 2009
            if equinox < pd.Timestamp("2009-01-01"):
                continue
            if equinox.dayofweek == WEDNESDAY:
                citizens_holidays.append(day_before(equinox))

        expected_citizens_holidays = CitizensHolidaySilverWeek
        # NOTE: Tests of imported holidays should probably be on a separate test module.
        assert citizens_holidays == expected_citizens_holidays  # Sanity check

        respect_for_the_aged_days = RespectForTheAgedDay2003Onwards.dates(
            "2003-01-01", AutumnalEquinoxes[-1]
        )

        rftad_holidays = []
        for citizens_holiday in citizens_holidays:
            # the day before the citizen's holiday should be Respect for the Aged Day.
            respect_for_the_aged_day = day_before(citizens_holiday)
            rftad_holidays.append(respect_for_the_aged_day)
            # NOTE: Tests of imported holidays should probably be on a separate module.
            assert respect_for_the_aged_day in respect_for_the_aged_days  # Sanity check

        yield citizens_holidays + rftad_holidays

    @pytest.fixture  # Calendar-specific fixture
    def golden_week_holidays(self):
        """List of pd.Timestamp representing citizen's holidays in Golden Week
        prior to 2007.
        """
        # from 1997 to 2006 May 4 was an unnamed citizen's holiday between
        # Constitution Memorial Day and Children's Day
        start, end = "1997-01-01", "2007-01-01"
        consitution_memorial_days = ConstitutionMemorialDay.dates(start, end)
        childrens_days = ChildrensDay.dates(start, end)
        citizens_days = []
        for cm_day, childrens_day in zip(
            consitution_memorial_days, childrens_days, strict=False
        ):
            # if there is only one day between Constitution Memorial
            # Day and Children's Day, that day should be a holiday
            if childrens_day - cm_day != pd.Timedelta(days=2):
                continue
            citizens_days.append(cm_day + pd.Timedelta(days=1))
        yield citizens_days

    @pytest.fixture
    def regular_holidays_sample(self, silver_week_holidays, golden_week_holidays):
        by_year = [
            # 2012
            "2012-01-01",  # New Year's holiday
            "2012-01-02",  # New Year's holiday
            "2012-01-03",  # New Year's holiday
            "2012-01-09",  # Coming of Age Day
            # National Foundation Day was on a Saturday so it is ignored
            "2012-03-20",  # Vernal Equinox
            "2012-04-30",  # Showa Day Observed
            "2012-05-03",  # Constitution Memorial Day
            "2012-05-04",  # Greenery Day
            # Children's Day was on a Saturday so it is ignored
            "2012-07-16",  # Marine Day
            "2012-09-17",  # Respect for the Aged Day
            # The Autumnal Equinox was on a Saturday so it is ignored
            "2012-10-08",  # Health and Sports Day
            # Culture Day was on a Saturday so it is ignored
            "2012-11-23",  # Labor Thanksgiving Day
            "2012-12-24",  # Emperor Birthday Observed
            "2012-12-31",  # New Year's holiday
            #
            # 2020
            "2020-01-01",  # New Year's holiday
            "2020-01-02",  # New Year's holiday
            "2020-01-03",  # New Year's holiday
            "2020-01-13",  # Coming of Age Day
            "2020-02-11",  # National Foundation Day
            "2020-02-23",  # Emperor's Birthday
            "2020-02-24",  # Emperor's Birthday observed
            "2020-03-20",  # Vernal Equinox
            "2020-04-29",  # Showa Day
            "2020-05-03",  # Constitution Memorial Day
            "2020-05-04",  # Greenery Day
            "2020-05-05",  # Children's Day
            "2020-05-06",  # Constitution Memorial Day
            # observed
            "2020-07-23",  # Marine Day
            "2020-07-24",  # Sports Day
            "2020-08-10",  # Mountain Day
            "2020-09-21",  # Respect for the Aged Day
            "2020-09-22",  # Autumnal Equinox
            "2020-11-03",  # Culture Day
            "2020-11-23",  # Labor Thanksgiving Day
            "2020-12-31",  # New Year's holiday
            #
            # 2021
            "2021-01-01",  # New Year's Day
            "2021-01-02",  # Market Holiday
            "2021-01-03",  # Market Holiday
            "2021-01-11",  # Coming of Age Day
            "2021-02-11",  # National Foundation Day
            "2021-02-23",  # Emperor's Birthday
            "2021-03-20",  # Vernal Equinox
            "2021-04-29",  # Showa Day
            "2021-05-03",  # Constitution Memorial Day
            "2021-05-04",  # Greenery Day
            "2021-05-05",  # Children's Day
            "2021-07-22",  # Marine Day
            "2021-07-23",  # Sports Day
            "2021-08-08",  # Mountain Day
            "2021-08-09",  # Mountain Day observed
            "2021-09-20",  # Respect for the Aged Day
            "2021-09-23",  # Autumnal Equinox
            "2021-11-03",  # Culture Day
            "2021-11-23",  # Labor Thanksgiving Day
            "2021-12-31",  # Market Holiday
            #
            # 2022
            "2022-01-01",  # New Year's Day
            "2022-01-02",  # Market Holiday
            "2022-01-03",  # Market Holiday
            "2022-01-10",  # Coming of Age Day
            "2022-02-11",  # National Foundation Day #
            "2022-02-23",  # Emperor's Birthday
            "2022-03-21",  # Vernal Equinox
            "2022-04-29",  # Showa Day
            "2022-05-03",  # Constitution Memorial Day
            "2022-05-04",  # Greenery Day
            "2022-05-05",  # Children's Day
            "2022-07-18",  # Marine Day
            "2022-08-11",  # Mountain Day
            "2022-09-19",  # Respect for the Aged Day
            "2022-09-23",  # Autumnal Equinox
            "2022-10-10",  # Sports Day
            "2022-11-03",  # Culture Day
            "2022-11-23",  # Labor Thanksgiving Day
            "2022-12-31",  # Market Holiday
            #
            # 2023
            "2023-01-01",  # New Year's Day
            "2023-01-02",  # Market Holiday
            "2023-01-03",  # Market Holiday
            "2023-01-09",  # Coming of Age Day
            "2023-02-11",  # National Foundation Day #
            "2023-02-23",  # Emperor's Birthday
            "2023-03-21",  # Vernal Equinox
            "2023-04-29",  # Showa Day
            "2023-05-03",  # Constitution Memorial Day
            "2023-05-04",  # Greenery Day
            "2023-05-05",  # Children's Day
            "2023-07-17",  # Marine Day
            "2023-08-11",  # Mountain Day
            "2023-09-18",  # Respect for the Aged Day
            "2023-09-23",  # Autumnal Equinox
            "2023-10-09",  # Sports Day
            "2023-11-03",  # Culture Day
            "2023-11-23",  # Labor Thanksgiving Day
            "2023-12-31",  # Market Holiday
            #
            # 2024
            "2024-01-01",  # New Year's Day
            "2024-01-02",  # Market Holiday
            "2024-01-03",  # Market Holiday
            "2024-01-08",  # Coming of Age Day
            "2024-02-11",  # National Foundation Day
            "2024-02-12",  # National Foundation Day observed
            "2024-02-23",  # Emperor's Birthday
            "2024-03-20",  # Vernal Equinox
            "2024-04-29",  # Showa Day
            "2024-05-03",  # Constitution Memorial Day
            "2024-05-04",  # Greenery Day
            "2024-05-05",  # Children's Day
            "2024-05-06",  # Children's Day observed
            "2024-07-15",  # Marine Day
            "2024-08-11",  # Mountain Day
            "2024-08-12",  # Mountain Day observed
            "2024-09-16",  # Respect for the Aged Day
            "2024-09-22",  # Autumnal Equinox
            "2024-09-23",  # Autumnal Equinox observed
            "2024-10-14",  # Sports Day
            "2024-11-03",  # Culture Day
            "2024-11-04",  # Culture Day observed
            "2024-11-23",  # Labor Thanksgiving Day
            "2024-12-31",  # Market Holiday
            #
            # 2025
            "2025-01-01",  # New Year's Day
            "2025-01-02",  # Market Holiday
            "2025-01-03",  # Market Holiday
            "2025-01-13",  # Coming of Age Day
            "2025-02-11",  # National Foundation Day
            "2025-02-23",  # Emperor's Birthday
            "2025-02-24",  # Emperor's Birthday observed
            "2025-03-20",  # Vernal Equinox
            "2025-04-29",  # Showa Day
            "2025-05-03",  # Constitution Memorial Day
            "2025-05-04",  # Greenery Day
            "2025-05-05",  # Children's Day
            "2025-05-06",  # Greenery Day observed
            "2025-07-21",  # Marine Day
            "2025-08-11",  # Mountain Day
            "2025-09-15",  # Respect for the Aged Day
            "2025-09-23",  # Autumnal Equinox
            "2025-10-13",  # Sports Day
            "2025-11-03",  # Culture Day
            "2025-11-23",  # Labor Thanksgiving Day
            "2025-11-24",  # Labor Thanksgiving Day observed
            "2025-12-31",  # Market Holiday
            # Mountain Day 11th August, observed from 2016.
            "2016-08-11",
            "2019-08-12",  # Falls on Sunday, made up on Monday.
        ]

        others = silver_week_holidays + golden_week_holidays
        others = [day.strftime("%Y-%m-%d") for day in others]

        yield by_year + others

    @pytest.fixture
    def adhoc_holidays_sample(self):
        yield [
            # 2019
            "2019-04-30",  # Abdication Day
            "2019-05-01",  # Accession Day
            "2019-05-02",  # Citizen's Holiday
            "2019-10-22",  # Enthronment Ceremony
            "2020-10-01",  # equity trading system failure
        ]

    @pytest.fixture
    def non_holidays_sample(self):
        yield [
            "2015-08-11",  # Mountain Day not observed until to 2016.
        ]

    # Calendar-specific tests

    def test_golden_week_holidays(self):
        # NOTE: This tests imported holidays rather than the XTKS calendar.
        # This test should probably be on a separate test module (together with other
        # such tests here and the sanity checks within fixtures).

        # from 2007 onwards, Greenery Day was moved to May 4
        start, end = "2007-01-01", "2019-01-01"
        consitution_memorial_days = ConstitutionMemorialDay.dates(start, end)
        greenery_days = GreeneryDay2007Onwards.dates(start, end)
        childrens_days = ChildrensDay.dates(start, end)

        # In 2008, Greenery Day is on a Sunday, and Children's Day
        # is on a Monday, so Greenery Day should be observed on Tuesday
        #       May 2008
        # Su Mo Tu We Th Fr Sa
        #  4  5  6  7  8  9 10
        assert pd.Timestamp("2008-05-05") in childrens_days
        assert pd.Timestamp("2008-05-06") in greenery_days

        # In 2009, Consitution Memorial Day should be observed on Wednesday,
        # since it is the next weekday that is not a holiday
        #       May 2009
        # Su Mo Tu We Th Fr Sa
        #                 1  2
        #  3  4  5  6  7  8  9
        assert pd.Timestamp("2009-05-04") in greenery_days
        assert pd.Timestamp("2009-05-05") in childrens_days
        assert pd.Timestamp("2009-05-06") in consitution_memorial_days

        # In 2012, Children's Day should not be observed because it falls
        # on a Saturday
        #       May 2012
        # Su Mo Tu We Th Fr Sa
        #        1  2  3  4  5
        #  6  7  8  9 10 11 12
        assert pd.Timestamp("2012-05-03") in consitution_memorial_days
        assert pd.Timestamp("2012-05-04") in greenery_days

        # In 2013, May 3 and 6 should be a holiday
        #       May 2013
        # Su Mo Tu We Th Fr Sa
        #           1  2  3  4
        #  5  6  7  8  9 10 11
        assert pd.Timestamp("2013-05-03") in consitution_memorial_days
        assert pd.Timestamp("2013-05-06") in childrens_days

    def test_emperors_birthday(self):
        # NOTE: This tests imported holidays rather than the XTKS calendar.
        # This test should probably be on a separate test module (together with other
        # such tests here and the sanity checks within fixtures).

        # The Emperor's birthday should be celebrated every year except
        # for 2019
        expected_birthdays = EmperorAkihitoBirthday.dates("1990-01-01", "2020-01-01")

        for year in range(1990, 2019):
            birthday = pd.Timestamp(f"{year}-12-23")
            if birthday.dayofweek == SUNDAY:
                birthday += pd.Timedelta(days=1)
            assert birthday in expected_birthdays

        assert pd.Timestamp("2019-12-23") not in expected_birthdays



import pytest

from exchange_calendars.exchange_calendar_xsau import XSAUExchangeCalendar
from .test_exchange_calendar import ExchangeCalendarTestBase
from .test_utils import T


class TestXASUCalendar(ExchangeCalendarTestBase):
    @pytest.fixture(scope="class")
    def calendar_cls(self):
        yield XSAUExchangeCalendar

    @pytest.fixture
    def start_bound(self):
        yield T("2021-01-01")

    @pytest.fixture
    def end_bound(self):
        yield T("2029-12-31")

    @pytest.fixture
    def max_session_hours(self):
        yield 5

    @pytest.fixture
    def regular_holidays_sample(self):
        yield ["2023-02-22", "2023-09-23"]



import pytest
import pandas as pd

from exchange_calendars.exchange_calendar_xnze import XNZEExchangeCalendar
from .test_exchange_calendar import ExchangeCalendarTestBase


class TestXNZECalendar(ExchangeCalendarTestBase):
    @pytest.fixture(scope="class")
    def calendar_cls(self):
        yield XNZEExchangeCalendar

    @pytest.fixture
    def max_session_hours(self):
        # The New Zealand Exchange is open from 10:00 am to 4:45 pm.
        yield 6.75

    @pytest.fixture
    def regular_holidays_sample(self):
        yield [
            # 2014
            "2014-01-01",  # New Year's Day
            "2014-01-02",  # Day after New Year's Day
            "2014-02-06",  # Waitangi Day
            "2014-04-18",  # Good Friday
            "2014-04-21",  # Easter Monday
            "2014-04-25",  # Anzac Day
            "2014-06-02",  # Queen's Birthday
            "2014-10-27",  # Labour Day
            "2014-12-25",  # Christmas
            "2014-12-26",  # Boxing Day
            # Holidays falling on a weekend and subsequently made up.
            "2010-12-27",  # Christmas Day fell on Saturday
            "2010-12-28",  # Boxing Day fell on Sunday
            "2011-01-03",  # New Year's Day fell on Saturday
            "2011-01-04",  # Day after New Year's Day (also a holdiay) fell on Sunday
            "2011-12-27",  # Christmas Day fell on Sunday
            "2012-01-02",  # Day after New Year's Day (a holiday) fell on Monday 2nd...
            "2012-01-03",  # but as New Year's Day fell on Sunday, made up Tuesday 3rd
            "2016-02-08",  # Waitangi Day fell on Saturday (2016 first year made up).
            "2015-04-27",  # Aztec Day fell on Saturday (2015 first year made up).
        ]

    @pytest.fixture
    def adhoc_holidays_sample(self):
        yield [
            "2022-06-24",
            "2035-06-29",
            "2049-06-25",  # Mataraki Day (a small selection)
            "2022-09-26",  # National day of mourning for the queen
        ]

    @pytest.fixture
    def non_holidays_sample(self):
        yield [
            # In 2012 New Year's Day fell on Sunday such that Day after New Year's Day
            # (also a holiday), fell on Monday 2nd. New Year's Day was therefore made up
            # on Tuesday 3rd. Make sure nothing being made up on Wednesday 4th...
            "2012-01-04",
            #
            # Prior to 2015, following holidays are not made up if fall on weekend.
            "2010-02-08",  # Monday that followed Waitangi Day that fell on Saturday.
            "2010-04-26",  # Monday that followed Anzac Day that fell on Sunday.
        ]

    @pytest.fixture
    def early_closes_sample(self):
        yield [
            # # 2011 was first observance of early closes on the trading day preceeding
            # Christmas Day and New Year's Day
            "2011-12-23",  # NB Christmas Day was a Sunday, so Friday an early close
            "2011-12-30",  # NB New Year's Day was a Sunday, so Friday an early close
            "2014-12-24",  # Christmas Eve
            "2014-12-31",  # New Year's Eve
        ]

    @pytest.fixture
    def early_closes_sample_time(self):
        yield pd.Timedelta(hours=12, minutes=45)

    @pytest.fixture
    def non_early_closes_sample(self):
        yield [
            "2010-12-24",  # Prior to 2011 Christmas Eve is not an early close.
            "2010-12-31",  # Prior to 2011 New Year's Eve is not an early close.
        ]

    @pytest.fixture
    def non_early_closes_sample_time(self):
        yield pd.Timedelta(hours=16, minutes=45)



from functools import partial

import click
import pandas as pd

from exchange_calendars import get_calendar
from exchange_calendars.calendar_helpers import UTC
from exchange_calendars.calendar_utils import default_calendar_names


class TimestampType(click.ParamType):
    """A click parameter that parses the value into pandas.Timestamp objects.

    Parameters
    ----------
    tz : timezone-coercable, optional
        The timezone to parse the string as.
        By default the timezone will be infered from the string or naiive.
    """

    def __init__(self, tz=None):
        self.tz = tz

    @property
    def name(self):
        return "TIMESTAMP"

    def convert(self, value, param, ctx):
        try:
            return pd.Timestamp(value, tz=self.tz)
        except ValueError:
            self.fail(
                f"{value} is not a valid timestamp.",
                param,
                ctx,
            )


def check_holidays(
    holiday_key_path,
    calendar_column,
    holiday_column,
    date_format,
    delimiter,
    strip_x_from_cal_name,
    answer_key_calendar_name,
    min_date,
    calendars,
):
    data = pd.read_csv(
        holiday_key_path,
        sep=delimiter,
        parse_dates=[holiday_column],
        date_parser=partial(pd.to_datetime, format=date_format, utc=True),
    )

    for calendar_name in default_calendar_names:
        if calendars and calendar_name not in calendars:
            continue

        cal = get_calendar(calendar_name)

        if answer_key_calendar_name:
            csv_cal_code = answer_key_calendar_name
        elif strip_x_from_cal_name:
            # Convert the calendar name to the format expected in the CSV.
            csv_cal_code = calendar_name.lstrip("X")
        else:
            csv_cal_code = calendar_name

        holidays = set(data[holiday_column][data[calendar_column] == csv_cal_code])

        if holidays:
            start = max(cal.first_session, min(holidays), min_date)
            end = min(cal.last_session, max(holidays))

            _check_range(start, end, holidays, cal, calendar_name)
        else:
            click.secho(
                f"No holidays found for {calendar_name} in the holiday key.",
                fg="yellow",
            )

        click.echo()


def _check_range(start, end, holidays, cal, calendar_name):
    msg = (
        f"Checking {len(holidays)} holidays on the {type(cal).__name__}"
        f" ({calendar_name}) calendar between {start.date()} and {end.date()}."
    )
    click.echo(msg)

    dates = pd.date_range(start, end, freq="B")

    expected_sessions = dates.difference(holidays)
    actual_sessions = cal.sessions_in_range(start, end)

    unexpected_sessions = actual_sessions.difference(expected_sessions)
    unexpected_holidays = expected_sessions.difference(actual_sessions)

    if unexpected_sessions.size:
        click.echo(
            "\nThese dates are holidays in the given key, but trading days in "
            f"the {calendar_name} exchange calendar:"
        )
        for session in unexpected_sessions:
            click.secho(str(session.date()), fg="red")

    if unexpected_holidays.size:
        click.echo(
            "\nThese dates are trading days in the given key, but holidays in "
            f"the {calendar_name} exchange calendar:"
        )
        for holiday in unexpected_holidays:
            click.secho(str(holiday.date()), fg="red")

    if not unexpected_sessions.size and not unexpected_holidays.size:
        click.secho(
            f"{calendar_name} calendar matches the holiday key.",
            fg="green",
        )


@click.command(
    help=(
        "Checks exchange calendars against a holiday key."
        " HOLIDAY_KEY_PATH is the path to a CSV file containing the key,"
        " with a calendar name column and a holiday date column."
    )
)
@click.option(
    "--min-date",
    default=pd.Timestamp("2002-01-01"),
    type=TimestampType(tz=UTC),
    help="Start the holiday comparison at this date.",
    show_default=True,
)
@click.option(
    "--calendar",
    "calendars",
    multiple=True,
    type=click.Choice(default_calendar_names),
    help=(
        "The name of a calendar to check. Specify multiple times to"
        " check multiple calendars. Default is to check all calendars."
    ),
)
@click.option("--calendar-column", help="The calendar name column in the CSV.")
@click.option("--holiday-column", help="The holiday date column in the CSV.")
@click.option(
    "--date-format",
    default="%Y%m%d",
    help="strftime format string for parsing the date column.",
    show_default=True,
)
@click.option(
    "--delimiter",
    default=",",
    help="Delimiter to use when parsing the key.",
    show_default=True,
)
@click.option(
    "--strip-x-from-cal-name/--no-strip-x-from-cal-name",
    help='E.g. "XNYS" is looked up in the key as "NYS".',
    show_default=True,
)
@click.option(
    "--answer-key-calendar-name",
    default=None,
    help=(
        "Use this if a calendar is identified differently in the answer key."
        " Not compatible with passing in multiple calendars in --calendars."
    ),
    show_default=False,
)
@click.argument("holiday_key_path", type=click.Path(exists=True))
def main(
    holiday_key_path,
    min_date,
    calendars,
    calendar_column,
    holiday_column,
    date_format,
    delimiter,
    strip_x_from_cal_name,
    answer_key_calendar_name,
):
    check_holidays(
        holiday_key_path,
        calendar_column,
        holiday_column,
        date_format,
        delimiter,
        strip_x_from_cal_name,
        answer_key_calendar_name,
        min_date,
        calendars,
    )


if __name__ == "__main__":
    main()



import pytest

from exchange_calendars.exchange_calendar_aixk import AIXKExchangeCalendar
from .test_exchange_calendar import ExchangeCalendarTestBase
from .test_utils import T


class TestAIXKCalendar(ExchangeCalendarTestBase):
    @pytest.fixture(scope="class")
    def calendar_cls(self):
        yield AIXKExchangeCalendar

    @pytest.fixture
    def max_session_hours(self):
        yield 6

    @pytest.fixture
    def start_bound(self):
        yield T("2017-01-01")

    @pytest.fixture
    def regular_holidays_sample(self):
        yield [
            # 2013
            "2013-12-01",  # First President Day
            # 2021
            "2021-01-01",  # New Year's Day
            "2021-01-07",  # Orthodox Christmas Day
            "2021-03-08",  # International Women's Day
            "2021-03-22",  # Nauryz Holiday
            "2021-03-23",  # Nauryz Holiday
            "2021-03-24",  # Nauryz Holiday
            "2021-05-03",  # Kazakhstan People Solidarity Day
            "2021-05-07",  # Defender's Day
            "2021-05-10",  # Victory Day Holiday
            "2021-07-06",  # Capital City Day
            "2021-07-20",  # Kurban Ait Holiday
            "2021-08-30",  # Constitution Day
            "2021-12-01",  # First President Day
            "2021-12-16",  # Independence Day
            "2021-12-17",  # Independence Day Holiday
            #
            # Holiday's made up when fall on weekend.
            # Last day of Nauryz on Saturday, 23th
            "2019-03-21",
            "2019-03-22",
            "2019-03-25",
            # First day of Nauryz on Sunday
            "2020-03-22",
            "2020-03-23",
            "2020-03-24",
            # Women's day on Sunday, Mar 8th
            "2020-03-09",
            # Capital day on Sunday, Jul 7th
            "2019-07-08",
        ]

    @pytest.fixture
    def adhoc_holidays_sample(self):
        # Bridge Days
        yield [
            "2018-03-09",  # between Women's day - Weekend
            "2018-04-30",  # between Weekend - Kazakhstan People Solidarity Day
            "2018-05-08",  # between Defender's Day - Victory Day
            "2018-08-31",  # between Constitution Day - Weekend
            "2018-12-31",  # between New Year's Eve - New Year's day
            "2019-05-10",  # between Victory Day - Weekend
            "2020-01-03",  # between New Year's day - Weekend
            "2020-12-18",  # between Independence day - Weekend
            "2021-06-05",  # between Weekend - Capital City day
            "2022-03-07",  # between Weekend - Women's day
        ]

    @pytest.fixture
    def non_holidays_sample(self):
        yield [
            "2022-12-01",  # First President Day
        ]



import pytest

from exchange_calendars.exchange_calendar_xphs import XPHSExchangeCalendar
from .test_exchange_calendar import ExchangeCalendarTestBase


class TestXPHSCalendar(ExchangeCalendarTestBase):
    @pytest.fixture(scope="class")
    def calendar_cls(self):
        yield XPHSExchangeCalendar

    @pytest.fixture
    def max_session_hours(self):
        # The XPHS is open from 9:30AM to 3:30PM
        yield 6

    @pytest.fixture
    def regular_holidays_sample(self):
        yield [
            # 2019 +
            "2019-01-01",  # New Year's Day
            "2019-02-05",  # Chinese New Year
            "2019-04-09",  # Araw Ng Kagitingan
            "2019-04-18",  # Maundy Thursday
            "2019-05-01",  # Labour Day
            "2019-06-05",  # Eid al-Fitr
            "2019-06-12",  # Independence Day
            "2019-08-12",  # Eid al-Adha
            "2019-08-21",  # Ninoy Aquino Day
            "2019-08-26",  # National Heroes Day
            "2017-09-01",  # Eid al-Adha
            "2019-11-01",  # All Saint's Day
            "2018-11-30",  # Bonifacio Day
            "2019-12-24",  # Christmas Eve
            "2019-12-25",  # Christmas
            "2019-12-30",  # Rizal Day
            "2019-12-31",  # New Year's Eve
            #
            # National Heroes' Day, last Monday of every August.
            "2019-08-26",
            "2018-08-27",
            "2017-08-28",
            "2016-08-29",
            "2015-08-31",
            "2014-08-25",
            "2013-08-26",
            "2012-08-27",
            "2011-08-29",
        ]

    @pytest.fixture
    def adhoc_holidays_sample(self):
        yield [
            # Holidays prior to 2011 are hard-coded for convenience. Below is a subset
            # of these holidays (those for 2008) to verify they are being recognised.
            "2008-01-01",
            "2008-02-25",
            "2008-03-20",
            "2008-03-21",
            "2008-04-07",
            "2008-05-01",
            "2008-06-09",
            "2008-08-18",
            "2008-08-25",
            "2008-10-01",
            "2008-12-01",
            "2008-12-25",
            "2008-12-26",
            "2008-12-29",
            "2008-12-30",
            "2008-12-31",
        ]

    @pytest.fixture
    def non_holidays_sample(self):
        yield [
            # Holidays that fall on a weekend and are not made up. Ensure surrounding
            # days are not holidays.
            # New Year's Day on Sunday, Jan 1st.
            #  2011-12-30 is Rizal Day
            "2011-12-29",
            "2012-01-02",
            # Chinese New Year's on Saturday, Jan 28th.
            "2017-01-27",
            "2017-01-30",
            # Araw Ng Kagitingan on Sunday, Apr 9th.
            "2017-04-07",
            "2017-04-10",
            # Labour Day on Saturday, May 1st.
            "2016-04-29",
            "2016-05-02",
            # Independence Day on Sunday, Jun 12th.
            "2016-06-10",
            "2016-06-13",
            # Ninoy Aquino Day on Sunday, Aug 21st.
            "2016-08-19",
            "2016-08-22",
            # All Saint's Day on Sunday, Nov 1st.
            "2015-10-30",
            "2015-11-02",
            # Bonifacio Day on Saturday, Nov 30th.
            "2019-11-29",
            "2019-12-02",
            # Christmas Eve + Christmas Day on weekend.
            "2011-12-23",
            "2011-12-26",
            # Rizal Day + New Year's Eve on weekend.
            #  2018-01-01 and 2018-01-02 are both holidays.
            "2017-12-29",
            "2018-01-03",
        ]



from datetime import time
from zoneinfo import ZoneInfo

from pandas.tseries.holiday import (
    Holiday,
    GoodFriday,
    EasterMonday,
)

from .common_holidays import (
    new_years_day,
    assumption_day,
    christmas_eve,
    christmas,
    boxing_day,
    new_years_eve,
)
from .exchange_calendar import HolidayCalendar, ExchangeCalendar

NewYearsDay = new_years_day()
NewYearHoliday = Holiday("New Year Holiday", month=1, day=2, start_date="2017")
NewYearHolidayBefore2013 = Holiday("New Year Holiday", month=1, day=2, end_date="2013")
PreserenDay = Holiday("Prešeren Day", month=2, day=8)
ResistanceDay = Holiday("Resistance Day", month=4, day=27)
LabourDay = Holiday("Labour Day", month=5, day=1)
LabourDay2 = Holiday("Labour Day (Second Day)", month=5, day=2)
StatehoodDay = Holiday("Statehood Day", month=6, day=25)
AssumptionDay = assumption_day()
ReformationDay = Holiday("Reformation Day", month=10, day=31)
AllSaintsDay = Holiday("All Saints' Day", month=11, day=1)
ChristmasEve = christmas_eve()
Christmas = christmas()
IndependenceUnityDay = Holiday("Independence and Unity Day", month=12, day=26)
BoxingDay = boxing_day()
NewYearsEve = new_years_eve()


class XLJUExchangeCalendar(ExchangeCalendar):
    """
    Calendar for the Ljubljana Stock Exchange (Slovenia).
    https://ljse.si/en/non-trading-days/110

    Open Time: 9:15 AM
    Close Time: 3:15 PM

    Regularly-Observed Holidays:
      - New Year's Day (Jan 1)
      - New Year Holiday (Jan 2)
      - Prešeren Day (Feb 8)
      - Good Friday
      - Easter Monday
      - Resistance Day (Apr 27)
      - Labour Day (May 1 & 2)
      - National Holiday (Jun 25)
      - Assumption Day (Aug 15)
      - Reformation Day (Oct 31)
      - All Saints' Day (Nov 1)
      - Christmas Eve (Dec 24)
      - Christmas Day (Dec 25)
      - Independence and Unity Day (Dec 26)
      - New Years Eve (Dec 31)
    """

    name = "XLJU"
    tz = ZoneInfo("Europe/Ljubljana")
    open_times = ((None, time(9, 15)),)
    close_times = ((None, time(15, 15)),)

    @property
    def regular_holidays(self):
        return HolidayCalendar(
            [
                NewYearsDay,
                NewYearHoliday,
                NewYearHolidayBefore2013,
                PreserenDay,
                GoodFriday,
                EasterMonday,
                ResistanceDay,
                LabourDay,
                LabourDay2,
                StatehoodDay,
                AssumptionDay,
                ReformationDay,
                AllSaintsDay,
                ChristmasEve,
                Christmas,
                IndependenceUnityDay,
                NewYearsEve,
            ]
        )

    @property
    def adhoc_holidays(self):
        return [
            "2023-08-14",  # Day off work due to the floods
            "2017-02-02",
            "2017-02-03",
        ]



from pathlib import Path
import requests

import pandas as pd

# Precomputed/adhoc KRX holidays can be checked here
# http://open.krx.co.kr/contents/MKD/01/0110/01100305/MKD01100305.jsp
# http://global.krx.co.kr/contents/GLB/05/0501/0501110000/GLB0501110000.jsp

download_holidays_as_dict_oldest_year_available = 1975


def download_krx_holidays_as_dict(year=None, page_first_call=False):
    now = pd.Timestamp.now("Asia/Seoul")

    if year is None:
        year = now.year

    if year < download_holidays_as_dict_oldest_year_available:
        raise ValueError(
            "Year cannot be older than"
            f" {download_holidays_as_dict_oldest_year_available} but {year} given"
        )

    def generate_otp():
        headers = {
            "Accept": "*/*",
            "Accept-Encoding": "gzip, deflate",
            "Host": "global.krx.co.kr",
            "Referer": "http://global.krx.co.kr/contents/GLB/05/0501/0501110000/GLB0501110000.jsp",
            "User-Agent": "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/88.0.4324.150 Safari/537.36",  # noqa: E501
            "X-Requested-With": "XMLHttpRequest",
        }
        params = {
            "bld": "GLB/05/0501/0501110000/glb0501110000_01",
            "name": "form",
            "_": str(int(now.timestamp() * 1000)),
        }
        response = requests.get(
            "http://global.krx.co.kr/contents/COM/GenerateOTP.jspx",
            headers=headers,
            params=params,
        )
        return response.content

    code = generate_otp()
    headers = {
        "Accept": "application/json, text/javascript, */*; q=0.01",
        "Accept-Encoding": "gzip, deflate",
        "Host": "global.krx.co.kr",
        "Origin": "http://global.krx.co.kr",
        "Referer": "http://global.krx.co.kr/contents/GLB/05/0501/0501110000/GLB0501110000.jsp",
        "User-Agent": "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/88.0.4324.150 Safari/537.36",  # noqa: E501
        "X-Requested-With": "XMLHttpRequest",
    }
    data = {
        "search_bas_yy": str(year),
        "gridTp": "KRX",
        "pagePath": "/contents/GLB/05/0501/0501110000/GLB0501110000.jsp",
        "code": code,
    }
    if page_first_call:
        data["pageFirstCall"] = "Y"
    response = requests.post(
        "http://global.krx.co.kr/contents/GLB/99/GLB99000001.jspx",
        headers=headers,
        data=data,
    )
    body = response.json()
    return body  # noqa: RET504


def get_precomputed_krx_holidays_online(from_year=None, to_year=None):
    if from_year is None:
        from_year = download_holidays_as_dict_oldest_year_available
    if to_year is None:
        now = pd.Timestamp.now("Asia/Seoul")
        to_year = now.year
    years = range(from_year, to_year + 1)
    precomputed_holidays = []
    for i, year in enumerate(years):
        page_first_call = i == 0
        result = download_krx_holidays_as_dict(year, page_first_call=page_first_call)
        for item in result["block1"]:
            holiday = item["calnd_dd"]
            precomputed_holidays.append(holiday)
    return pd.to_datetime(precomputed_holidays)


def update_dupmed_precomputed_krx_holidays():
    xkrx_holidays_py = Path(__file__).parent / "../exchange_calendars/xkrx_holidays.py"
    with xkrx_holidays_py.open("r") as f:
        lines = list(f)
    start_line = "dumped_precomputed_krx_holidays = pd.DatetimeIndex("
    end_line = ")"
    start_line_index = 0
    end_line_index = 0
    for i, line in enumerate(lines):
        if line.startswith(start_line):
            start_line_index = i + 2
        elif start_line_index > 0 and line.startswith(end_line):
            end_line_index = i - 2
            break
    if start_line_index > 0 and end_line_index > 0:
        precomputed_holidays = get_precomputed_krx_holidays_online()
        precomputed_lines = [
            '        "{}",\n'.format(pd.Timestamp(holiday).strftime("%Y-%m-%d"))
            for holiday in precomputed_holidays
        ]
        replaced_lines = (
            lines[:start_line_index] + precomputed_lines + lines[end_line_index + 1 :]
        )
        with xkrx_holidays_py.open("w") as f:
            f.write("".join(replaced_lines))


if __name__ == "__main__":
    update_dupmed_precomputed_krx_holidays()



from datetime import time
from zoneinfo import ZoneInfo

from pandas.tseries.holiday import EasterMonday, GoodFriday

from .common_holidays import (
    boxing_day,
    christmas,
    christmas_eve,
    european_labour_day,
    new_years_day,
    new_years_eve,
)
from .exchange_calendar import ExchangeCalendar, HolidayCalendar

# Regular Holidays
# ----------------
NewYearsDay = new_years_day()

EuropeanLabourDay = european_labour_day()

ChristmasEve = christmas_eve()

Christmas = christmas()

BoxingDay = boxing_day()

NewYearsEve = new_years_eve()


class XEURExchangeCalendar(ExchangeCalendar):
    """
    Exchange calendar for Eurex (XEUR).
    https://www.eurex.com/ex-en/trade/trading-calendar/holiday-regulations

    Open Time: 8:00 AM, CET
    Close Time: 22:00 PM, CET

    Regularly-Observed Holidays:
    - New Years Day
    - Good Friday
    - Easter Monday
    - Labour Day
    - Christmas Eve
    - Christmas Day
    - Boxing Day
    - New Year's Eve
    """

    name = "XEUR"

    tz = ZoneInfo("Europe/Berlin")

    open_times = ((None, time(8)),)

    close_times = ((None, time(22, 0)),)

    @property
    def regular_holidays(self):
        return HolidayCalendar(
            [
                NewYearsDay,
                GoodFriday,
                EasterMonday,
                EuropeanLabourDay,
                ChristmasEve,
                Christmas,
                BoxingDay,
                NewYearsEve,
            ]
        )



import pytest
from exchange_calendars.exchange_calendar_xbra import XBRAExchangeCalendar
from .test_exchange_calendar import ExchangeCalendarTestBase


class TestXBRAExchangeCalendar(ExchangeCalendarTestBase):
    @pytest.fixture(scope="class")
    def calendar_cls(self):
        yield XBRAExchangeCalendar

    @pytest.fixture
    def max_session_hours(self):
        # The XBRA is open from 11:00 am to 3:30 pm.
        yield 4.5

    @pytest.fixture
    def regular_holidays_sample(self):
        yield [
            "2025-01-01",  # New Year's Day
            "2025-01-06",  # Epiphany
            "2025-04-18",  # Good Friday
            "2025-04-21",  # Easter Monday
            "2025-05-01",  # Labour Day
            "2025-05-08",  # Victory Day
            "2025-07-05",  # Saints Cyril and Methodius Day
            "2025-08-29",  # Slovak National Uprising Anniversary
            "2025-09-15",  # Day of Our Lady of Sorrows
            "2025-11-01",  # All Saints' Day
            "2025-11-17",  # Struggle for Freedom and Democracy Day
            "2025-12-24",  # Christmas Eve
            "2025-12-25",  # Christmas Day
            "2025-12-26",  # Boxing Day
            "2025-12-31",  # New Year's Eve
        ]

    @pytest.fixture
    def non_holidays_sample(self):
        yield [
            "2025-01-02",
            "2025-01-07",
            "2025-04-17",
            "2025-04-22",
            "2025-05-02",
            "2025-05-09",
            "2025-08-28",
            "2025-09-02",
            "2025-09-16",
            "2025-11-18",
            "2025-12-23",
        ]



from datetime import time
from zoneinfo import ZoneInfo

import pandas as pd
from pandas.tseries.holiday import (
    Holiday,
    EasterMonday,
    GoodFriday,
)
from .common_holidays import (
    new_years_day,
    epiphany,
    european_labour_day,
    all_saints_day,
    christmas_eve,
    christmas,
    boxing_day,
    new_years_eve,
)
from .exchange_calendar import HolidayCalendar, ExchangeCalendar

NewYearsDay = new_years_day()
Epiphany = epiphany()
LabourDay = european_labour_day()
VictoryDay = Holiday("Victory Day", month=5, day=8)
SaintsCyrilMethodius = Holiday("Saints Cyril and Methodius Day", month=7, day=5)
SlovakNationalUprising = Holiday(
    "Slovak National Uprising Anniversary", month=8, day=29
)
ConstitutionDay = Holiday("Constitution Day", month=9, day=1, end_date="2024")
VirginMary = Holiday("Day of Our Lady of Sorrows", month=9, day=15)
AllSaintsDay = all_saints_day()
StruggleForFreedom = Holiday("Struggle for Freedom and Democracy Day", month=11, day=17)
ChristmasEve = christmas_eve()
Christmas = christmas()
BoxingDay = boxing_day()
NewYearsEve = new_years_eve(start_date="2016")


class XBRAExchangeCalendar(ExchangeCalendar):
    """
    Calendar for the Bratislava Stock Exchange (Slovakia).
    https://www.bsse.sk/bcpb/en/trading-days/

    Open Time: 11:00 AM
    Close Time: 3:30 PM

    Regularly-Observed Holidays:
      - New Year's Day (Jan 1)
      - Epiphany (Jan 6)
      - Good Friday
      - Easter Monday
      - Labour Day (May 1)
      - Victory Day (May 8)
      - Saints Cyril and Methodius Day (Jul 5)
      - Slovak National Uprising Anniversary (Aug 29)
      - Constitution Day (Sep 1)
      - Day of Our Lady of Sorrows (Sep 15)
      - All Saints' Day (Nov 1)
      - Struggle for Freedom and Democracy Day (Nov 17)
      - Christmas Eve (Dec 24)
      - Christmas Day (Dec 25)
      - Boxing Day (Dec 26)
      - New Year's Eve (Dec 31)
    """

    name = "XBRA"
    tz = ZoneInfo("Europe/Bratislava")
    open_times = ((None, time(11, 00)),)
    close_times = ((None, time(15, 30)),)

    @property
    def regular_holidays(self):
        return HolidayCalendar(
            [
                NewYearsDay,
                Epiphany,
                GoodFriday,
                EasterMonday,
                LabourDay,
                VictoryDay,
                SaintsCyrilMethodius,
                SlovakNationalUprising,
                ConstitutionDay,
                VirginMary,
                AllSaintsDay,
                StruggleForFreedom,
                ChristmasEve,
                Christmas,
                BoxingDay,
                NewYearsEve,
            ]
        )

    @property
    def adhoc_holidays(self):
        return [
            pd.Timestamp("2018-01-02"),  # Independent Slovakia, 25th anniversary
            pd.Timestamp(
                "2018-10-30"
            ),  # 100th anniversary of the adoption of the Declaration of
            # the Slovak Nation
        ]



#
# Copyright 2018 Quantopian, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from datetime import time
from zoneinfo import ZoneInfo

import pandas as pd
from pandas.tseries.holiday import (
    MO,
    DateOffset,
    EasterMonday,
    GoodFriday,
    Holiday,
    previous_friday,
    weekend_to_monday,
)

from .common_holidays import (
    boxing_day,
    christmas,
    new_years_day,
    weekend_boxing_day,
    weekend_christmas,
)
from .exchange_calendar import HolidayCalendar, ExchangeCalendar

# Regular Holidays
# ----------------
# New Year's Day
LSENewYearsDay = new_years_day(observance=weekend_to_monday)

# Early May bank holiday has two exceptions based on the 50th and 75th
# anniversary of VE-Day
# 1995-05-01 Early May bank holiday removed for VE-day 50th anniversary
# 2020-05-04 Early May bank holiday removed for VE-day 75th anniversary

# Early May bank holiday pre-1995
MayBank_pre_1995 = Holiday(
    "Early May Bank Holiday",
    month=5,
    offset=DateOffset(weekday=MO(1)),
    day=1,
    end_date=pd.Timestamp("1994-12-31"),
)

# Early May bank holiday post-1995 and pre-2020
MayBank_post_1995_pre_2020 = Holiday(
    "Early May Bank Holiday",
    month=5,
    offset=DateOffset(weekday=MO(1)),
    day=1,
    start_date=pd.Timestamp("1996-01-01"),
    end_date=pd.Timestamp("2019-12-31"),
)

# Early May bank holiday post 2020
MayBank_post_2020 = Holiday(
    "Early May Bank Holiday",
    month=5,
    offset=DateOffset(weekday=MO(1)),
    day=1,
    start_date=pd.Timestamp("2021-01-01"),
)


# Spring bank holiday has three exceptions based on the Golden, Diamond & Platinum Jubilee  # noqa: E501
# 2002-05-27 Spring bank holiday removed for Golden Jubilee
# 2012-05-28 Spring bank holiday removed for Diamond Jubilee
# 2022-05-30 Spring bank holiday moved to 2022-06-02 to be adjacent to the Platinum Jubilee day 2022-06-03  # noqa: E501

# Spring bank holiday
SpringBank_pre_2002 = Holiday(
    "Spring Bank Holiday",
    month=5,
    day=31,
    offset=DateOffset(weekday=MO(-1)),
    end_date=pd.Timestamp("2001-12-31"),
)

SpringBank_post_2002_pre_2012 = Holiday(
    "Spring Bank Holiday",
    month=5,
    day=31,
    offset=DateOffset(weekday=MO(-1)),
    start_date=pd.Timestamp("2003-01-01"),
    end_date=pd.Timestamp("2011-12-31"),
)

SpringBank_post_2012_pre_2022 = Holiday(
    "Spring Bank Holiday",
    month=5,
    day=31,
    offset=DateOffset(weekday=MO(-1)),
    start_date=pd.Timestamp("2013-01-01"),
    end_date=pd.Timestamp("2021-12-31"),
)

SpringBank_post_2022 = Holiday(
    "Spring Bank Holiday",
    month=5,
    day=31,
    offset=DateOffset(weekday=MO(-1)),
    start_date=pd.Timestamp("2023-01-01"),
)

# Summer bank holiday
SummerBank = Holiday(
    "Summer Bank Holiday",
    month=8,
    day=31,
    offset=DateOffset(weekday=MO(-1)),
)

Christmas = christmas()

WeekendChristmas = weekend_christmas()

BoxingDay = boxing_day()

WeekendBoxingDay = weekend_boxing_day()

# Early Closes
# ------------
# If Christmas Eve falls on a weekday, that day is a half day.
# If it falls on a weekend, the preceding Friday is a half day.
ChristmasEve = Holiday(
    "Christmas Eve",
    month=12,
    day=24,
    observance=previous_friday,
)
# New Year's eve (or the preceding Friday if it falls on a weekend)
# is a half day. Except for 1999-12-31, when the Queen declared a
# bank holiday.
NewYearsEvePre1999 = Holiday(
    "New Year's Eve",
    month=12,
    day=31,
    observance=previous_friday,
    end_date=pd.Timestamp("1999-01-01"),
)
NewYearsEvePost2000 = Holiday(
    "New Year's Eve",
    month=12,
    day=31,
    observance=previous_friday,
    start_date=pd.Timestamp("2000-01-01"),
)


class XLONExchangeCalendar(ExchangeCalendar):
    """
    Exchange calendar for the London Stock Exchange (XLON).

    Open Time: 8:00 AM, GMT
    Close Time: 4:30 PM, GMT

    Regularly-Observed Holidays:
    - New Years Day (observed on first business day on/after)
    - Good Friday
    - Easter Monday
    - Early May Bank Holiday (first Monday in May)
    - Spring Bank Holiday (last Monday in May)
    - Summer Bank Holiday (last Monday in May)
    - Christmas Day
    - Dec. 27th (if Christmas is on a weekend)
    - Boxing Day
    - Dec. 28th (if Boxing Day is on a weekend)

    Early Closes:
    - Christmas Eve
    - New Year's Eve
    """

    regular_early_close = time(12, 30)

    name = "XLON"

    tz = ZoneInfo("Europe/London")

    open_times = ((None, time(8)),)

    close_times = ((None, time(16, 30)),)

    @property
    def regular_holidays(self):
        return HolidayCalendar(
            [
                LSENewYearsDay,
                GoodFriday,
                EasterMonday,
                MayBank_pre_1995,
                MayBank_post_1995_pre_2020,
                MayBank_post_2020,
                SpringBank_pre_2002,
                SpringBank_post_2002_pre_2012,
                SpringBank_post_2012_pre_2022,
                SpringBank_post_2022,
                SummerBank,
                Christmas,
                WeekendChristmas,
                BoxingDay,
                WeekendBoxingDay,
            ]
        )

    @property
    def adhoc_holidays(self):
        return [
            # VE-Day Anniversary
            pd.Timestamp("1995-05-08"),  # 50th Anniversary
            pd.Timestamp("2020-05-08"),  # 75th Anniversary
            # Queen Elizabeth II Jubilees
            # Silver Jubilee
            pd.Timestamp("1977-06-07"),
            # Golden Jubilee
            pd.Timestamp("2002-06-03"),
            pd.Timestamp("2002-06-04"),
            # Diamond Jubilee
            pd.Timestamp("2012-06-04"),
            pd.Timestamp("2012-06-05"),
            # Platinum Jubilee (moved Spring bank holiday followed by Jubilee day)
            pd.Timestamp("2022-06-02"),
            pd.Timestamp("2022-06-03"),
            # Queen Elizabeth II Funeral
            pd.Timestamp("2022-09-19"),
            # Coronation of King Charles III
            pd.Timestamp("2023-05-08"),
            # Royal Weddings
            # Wedding Day of Princess Anne and Mark Phillips
            pd.Timestamp("1973-11-14"),
            # Wedding Day of Prince Charles and Diana Spencer
            pd.Timestamp("1981-07-29"),
            # Wedding Day of Prince William and Catherine Middleton
            pd.Timestamp("2011-04-29"),
            # Miscellaneous
            # Eve of 3rd Millenium A.D.
            pd.Timestamp("1999-12-31"),
        ]

    @property
    def special_closes(self):
        return [
            (
                self.regular_early_close,
                HolidayCalendar(
                    [
                        ChristmasEve,
                        NewYearsEvePre1999,
                        NewYearsEvePost2000,
                    ]
                ),
            )
        ]



#
# Copyright 2019 Quantopian, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import datetime
from zoneinfo import ZoneInfo

from pandas.tseries.holiday import Easter, EasterMonday, Holiday
from pandas.tseries.offsets import Day

from .common_holidays import (
    all_saints_day,
    christmas,
    christmas_eve,
    european_labour_day,
    new_years_day,
    new_years_eve,
    whit_monday,
)
from .exchange_calendar import THURSDAY, TUESDAY, HolidayCalendar, ExchangeCalendar


ONE_DAY = datetime.timedelta(1)


def bridge_mon(dt: datetime.datetime) -> datetime.datetime | None:
    """Define Monday as holiday if Tuesday is a holiday.

    If a holiday falls on a Tuesday an extra holiday is observed on Monday
    to bridge the weekend and the official holiday.
    """
    return dt - ONE_DAY if dt.weekday() == TUESDAY else None


def bridge_fri(dt: datetime.datetime) -> datetime.datetime | None:
    """Define Friday as holiday if Thrusday is a holiday.

    If a holiday falls on a Thursday an extra holiday is observed on Friday
    to bridge the weekend and the official holiday.
    """
    return dt + ONE_DAY if dt.weekday() == THURSDAY else None


NewYearsDayExtraMon = new_years_day(observance=bridge_mon)
NewYearsDayExtraFri = new_years_day(observance=bridge_fri)

NationalHoliday1 = Holiday("National Day", month=3, day=15)
NationalHoliday1ExtraMon = Holiday(
    "National Day extra Monday", month=3, day=15, observance=bridge_mon
)
NationalHoliday1ExtraFri = Holiday(
    "National Day extra Friday", month=3, day=15, observance=bridge_fri
)

# Need custom start year so can't use pandas GoodFriday
GoodFriday = Holiday(
    "Good Friday", month=1, day=1, offset=[Easter(), Day(-2)], start_date="2012"
)

LabourDayExtraMon = european_labour_day(observance=bridge_mon)
LabourDayExtraFri = european_labour_day(observance=bridge_fri)

WhitMonday = whit_monday()

StStephensDay = Holiday("St. Stephen's Day", month=8, day=20)
StStephensDayExtraMon = Holiday(
    "St. Stephen's Day extra Monday", month=8, day=20, observance=bridge_mon
)
StStephensDayExtraFri = Holiday(
    "St. Stephen's Day extra Friday", month=8, day=20, observance=bridge_fri
)

NationalHoliday2 = Holiday("National Day", month=10, day=23)
NationalHoliday2ExtraMon = Holiday(
    "National Day extra Monday", month=10, day=23, observance=bridge_mon
)
NationalHoliday2ExtraFri = Holiday(
    "National Day extra Friday", month=10, day=23, observance=bridge_fri
)

AllSaintsDayExtraMon = all_saints_day(observance=bridge_mon)
AllSaintsDayExtraFri = all_saints_day(observance=bridge_fri)

# Christmas Eve does not follow the four day weekend rule
ChristmasEve = christmas_eve()

ChristmasDay = christmas()

# XBUD always has a holiday for the second day of Christmas (26th),
# but starting in 2013 if the 26th falls on a Thursday then the
# 27th (Friday) is also taken off
SecondDayOfChristmas = Holiday("Second Day of Christmas", month=12, day=26)
SecondDayOfChristmasExtraFri = Holiday(
    "Second Day of Christmas extra Friday",
    month=12,
    day=26,
    start_date="2013",
    observance=bridge_fri,
)
# SecondDayOfChristmas unnecessary to bridge_mon as if second day of christmas
# is a tuesday then monday will already be a holiday (Chirstmas Day)


# Starting in 2011, New Year's Eve is observed as a holiday every year.
# In some cases pre-2011, the 31st becomes a holiday due to the four day
# weekend rule (when Jan 1 falls on a Tuesday).
# Also, when NYE starts being observed as a holiday it does NOT follow
# the four day weekend rule (no 30ths are holidays)
NewYearsEve = new_years_eve(start_date="2011")


class XBUDExchangeCalendar(ExchangeCalendar):
    """
    Exchange calendar for the Budapest Stock Exchange (XBUD).

    Open Time: 9:00 AM, CET
    Close Time: 5:00 PM, CET

    Regularly-Observed Holidays:
    - New Year's Day
    - National Holiday (Mar 15)
    - Good Friday
    - Easter Monday
    - Labour Day (May 1)
    - Whit Monday (50 days after Easter Sunday)
    - St. Stephen's Day (Aug 20)
    - National Holiday (Oct 23)
    - All Saint's Day (Nov 1)
    - Christmas Eve
    - Christmas Day
    - Second Day of Christmas (Dec 26)
    - New Year's Eve

    Early Closes:
    - None
    """

    name = "XBUD"

    tz = ZoneInfo("Europe/Budapest")

    open_times = ((None, datetime.time(9)),)

    close_times = ((None, datetime.time(17, 00)),)

    @property
    def regular_holidays(self):
        return HolidayCalendar(
            [
                new_years_day(),
                NewYearsDayExtraMon,
                NewYearsDayExtraFri,
                NationalHoliday1,
                NationalHoliday1ExtraMon,
                NationalHoliday1ExtraFri,
                GoodFriday,
                EasterMonday,
                european_labour_day(),
                LabourDayExtraMon,
                LabourDayExtraFri,
                WhitMonday,
                StStephensDay,
                StStephensDayExtraMon,
                StStephensDayExtraFri,
                NationalHoliday2,
                NationalHoliday2ExtraMon,
                NationalHoliday2ExtraFri,
                all_saints_day(),
                AllSaintsDayExtraMon,
                AllSaintsDayExtraFri,
                ChristmasEve,
                ChristmasDay,
                SecondDayOfChristmas,
                SecondDayOfChristmasExtraFri,
                NewYearsEve,
            ]
        )



import pytest
import pandas as pd

from exchange_calendars.exchange_calendar_xcbf import XCBFExchangeCalendar
from .test_exchange_calendar import ExchangeCalendarTestBase


class TestXCBFCalendar(ExchangeCalendarTestBase):
    @pytest.fixture(scope="class")
    def calendar_cls(self):
        yield XCBFExchangeCalendar

    @pytest.fixture
    def max_session_hours(self):
        yield 8

    @pytest.fixture
    def regular_holidays_sample(self):
        yield [
            # 2016
            "2016-01-01",  # new years: jan 1
            "2016-01-18",  # mlk: jan 18
            "2016-02-15",  # presidents: feb 15
            "2016-03-25",  # good friday: mar 25
            "2016-05-30",  # mem day: may 30
            "2016-07-04",  # independence day: july 4
            "2016-09-05",  # labor day: sep 5
            "2016-11-24",  # thanksgiving day: nov 24
            "2016-12-26",  # christmas (observed): dec 26
            "2017-01-02",  # new years (observed): jan 2 2017
            # 2022
            "2022-06-20",  # juneteenth (observed): jun 20 2022
        ]

    @pytest.fixture
    def adhoc_holidays_sample(self):
        yield [
            "1994-04-27",  # hurricane sandy: oct 29 2012, oct 30 2012
            "2004-06-11",  # national days of mourning:
            "2007-01-02",  # - apr 27 1994
            "2012-10-29",  # - june 11 2004
            "2012-10-30",  # - jan 2 2007
        ]

    @pytest.fixture
    def early_closes_sample(self):
        yield ["2016-11-25"]

    @pytest.fixture
    def early_closes_sample_time(self):
        yield pd.Timedelta(hours=12, minutes=15)



from datetime import time
from zoneinfo import ZoneInfo

import pandas as pd
from pandas.tseries.holiday import EasterMonday, GoodFriday, Holiday

from .common_holidays import (
    ascension_day,
    boxing_day,
    christmas_eve,
    new_years_day,
    new_years_eve,
)
from .exchange_calendar import HolidayCalendar, ExchangeCalendar

NewYearsDay = new_years_day()
AscensionDay = ascension_day()
ChristmasEve = christmas_eve()
BoxingDay = boxing_day()
NewYearsEve = new_years_eve()

IndependenceDay = Holiday("Independence Day", month=2, day=24)
SpringDay = Holiday("Spring Day", month=5, day=1)
VictoryDay = Holiday("Victory Day", month=6, day=23)
MidsummerDay = Holiday("Saint John's Day", month=6, day=24)
RestorationOfIndependence = Holiday("Independence Restoration Day", month=8, day=20)


class XTALExchangeCalendar(ExchangeCalendar):
    """
    Calendar for the Tallinn Stock Exchange (Estonia).
    https://nasdaqbaltic.com/statistics/en/calendar?holidays=1

    Open Time: 10:00 AM, EET (Eastern European Time)
    Close Time: 4:00 PM, EET (Eastern European Time)

    Regularly-Observed Holidays:
      - New Year's Day
      - Independence Day (Feb 24)
      - Good Friday
      - Easter Monday
      - Spring Day (May 1)
      - Ascension Day
      - Victory Day (Jun 23)
      - Midsummer Day (Jun 24)
      - Restoration of Independence (Aug 20)
      - Christmas Eve (Dec 24)
      - Boxing Day (Dec 26)
      - New Year's Eve (Dec 31)
    """

    name = "XTAL"
    tz = ZoneInfo("Europe/Tallinn")
    open_times = ((None, time(10, 0)),)
    close_times = ((None, time(16, 0)),)

    @property
    def regular_holidays(self):
        return HolidayCalendar(
            [
                NewYearsDay,
                IndependenceDay,
                GoodFriday,
                EasterMonday,
                SpringDay,
                AscensionDay,
                VictoryDay,
                MidsummerDay,
                RestorationOfIndependence,
                ChristmasEve,
                BoxingDay,
                NewYearsEve,
            ]
        )

    @property
    def adhoc_holidays(self):
        return [
            pd.Timestamp("2023-12-25"),  # Additional Day off
        ]



import pytest

from exchange_calendars.exchange_calendar_xhel import XHELExchangeCalendar
from .test_exchange_calendar import ExchangeCalendarTestBase


class TestXHELCalendar(ExchangeCalendarTestBase):
    @pytest.fixture(scope="class")
    def calendar_cls(self):
        yield XHELExchangeCalendar

    @pytest.fixture
    def max_session_hours(self):
        yield 8.5

    @pytest.fixture
    def regular_holidays_sample(self):
        yield [
            "2018-01-01",  # New Year's Day
            "2017-01-06",  # Epiphany
            "2018-03-30",  # Good Friday
            "2018-04-02",  # Easter Monday
            "2018-05-01",  # Labour Day
            "2018-05-10",  # Ascension Day
            "2018-06-22",  # Midsummer Eve
            "2018-12-06",  # Finland Independence Day
            "2018-12-24",  # Christmas Eve
            "2018-12-25",  # Christmas Day
            "2018-12-26",  # Boxing Day
            "2018-12-31",  # New Year's Ev
            # Midsummer Eve falls on Friday following June 18th.
            "2010-06-25",  # fell prior Friday June 18th
            "2017-06-23",  # fell prior Sunday
        ]

    @pytest.fixture
    def adhoc_holidays_sample(self):
        yield []

    @pytest.fixture
    def non_holidays_sample(self):
        yield [
            # Midsummer Eve falls on Friday following June 18th. In 2010, June 18th was
            # a Friday - check not a holiday (holiday is following Friday, June 25th).
            "2010-06-18",
            #
            # Holidays that fall on a weekend are not made up. Ensure surrounding
            # days are not holidays.
            # In 2018, the Epiphany fell on a Saturday, so the market should be
            # open on both the prior Friday and the following Monday.
            "2018-01-05",
            "2018-01-08",
            # In 2010, Labour Day fell on a Saturday, so the market should be
            # open on both the prior Friday and the following Monday.
            "2010-04-30",
            "2010-05-03",
            # In 2015, Finland Independence Day fell on a Sunday, so the market
            # should be open on both the prior Friday and the following Monday.
            "2015-12-04",
            "2015-12-07",
            # In 2010, Christmas fell on a Saturday, meaning Boxing Day fell on
            # a Sunday. The market should thus be open on the following Monday.
            "2010-12-27",
            # In 2017, New Year's Day fell on a Sunday, so the market should be
            # open on both the prior Friday and the following Monday.
            "2016-12-30",
            "2017-01-02",
        ]



import pytest

from exchange_calendars.exchange_calendar_xmos import XMOSExchangeCalendar
from .test_exchange_calendar import ExchangeCalendarTestBase


class TestXMOSCalendar(ExchangeCalendarTestBase):
    @pytest.fixture(scope="class")
    def calendar_cls(self):
        yield XMOSExchangeCalendar

    @pytest.fixture
    def max_session_hours(self):
        # The XMOS is open from 10:00AM to 6:45PM.
        yield 8.75

    @pytest.fixture
    def regular_holidays_sample(self):
        yield [
            # 2022
            "2022-03-07",  # Women's Day falls on a Tuesday.
            "2022-05-03",  # LabourDay Holiday extended to Tuesday.
            "2022-05-10",  # Victory Day Holiday extended to Tuesday.
            # 2020
            "2020-06-24",  # Non-Working Day
            "2020-07-01",  # Non-Working Day
            # 2019
            "2019-01-01",  # New Year's Day
            "2019-01-02",  # New Year's Holiday
            "2019-01-07",  # Orthodox Christmas
            "2018-02-23",  # Defender of the Fatherland
            "2019-03-08",  # Women's Day
            "2019-05-01",  # Labour Day
            "2019-05-09",  # Victory Day
            "2019-06-12",  # Day of Russia
            "2019-11-04",  # Unity Day
            "2019-12-31",  # New Year's Eve
            #
            # Holidays falling on a weekend and made up on following Monday.
            # New Year's Day on a Saturday and New Year's Holiday on a Sunday.
            "2011-01-03",
            # Orthodox Christmas (January 7th) on a Sunday.
            "2018-01-08",
            # Defender of the Fatherland Day (February 23rd) on a Saturday.
            "2008-02-25",
            # Women's Day (March 8th) on a Sunday.
            "2015-03-09",
            # Labour Day (May 1st) on a Sunday.
            "2016-05-02",
            # Victory Day (May 9th) on a Saturday.
            "2015-05-11",
            # Day of Russia (June 12th) on a Sunday.
            "2016-06-13",
            # Unity Day (November 4th) on a Sunday.
            "2018-11-05",
            #
            # "bridge days" where a Monday or Friday is made into a holiday
            # to fill gap between weekend and a holiday on Tuesday or Thursday.
            "2016-01-08",  # Orthodox Christmas falls on a Thursday.
            "2010-02-22",  # Defender of the Fatherland falls on a Tuesday.
            "2012-03-09",  # Women's Day falls on a Thursday.
            "2012-04-30",  # Labour Day falls on a Tuesday.
            "2017-05-08",  # Victory Day falls on a Tuesday.
            "2014-06-13",  # Day of Russia falls on a Thursday.
            "2010-11-05",  # Unity Day falls  on a Thursday.
        ]

    @pytest.fixture
    def non_holidays_sample(self):
        yield [
            # Holidays that fall on a weekend and are not made up. Ensure not made up.
            # New Year's Holiday on a Saturday, but the following Monday is a
            # trading day.
            "2016-01-04",
            "2021-01-04",
            "2022-01-03",
            # Orthodox Christmas on a Saturday, but the following Monday is a
            # trading day.
            "2017-01-09",
            "2023-01-09",
            # Orthodox Christmas (January 7th) on a Sunday, but
            # the following Monday is a trading day.
            "2024-01-08",
            # Defender of the Fatherland Day (February 23rd) on a Saturday, but
            # the following Monday is a trading day.
            "2019-02-25",
            # Victory Day (May 9th) on a Sunday, but
            # the following Monday is a trading day.
            "2021-05-10",
            # Day of Russia (June 12th) on a Saturday, but
            # the following Monday is a trading day.
            "2021-06-14",
        ]



import pytest
from exchange_calendars.exchange_calendar_xcys import XCYSExchangeCalendar
from .test_exchange_calendar import ExchangeCalendarTestBase


class TestXCYSExchangeCalendar(ExchangeCalendarTestBase):
    @pytest.fixture(scope="class")
    def calendar_cls(self):
        yield XCYSExchangeCalendar

    @pytest.fixture
    def max_session_hours(self):
        yield 6.5

    @pytest.fixture
    def regular_holidays_sample(self):
        yield [
            "2024-01-01",  # New Year's Day
            "2024-01-06",  # Epiphany
            "2024-03-18",  # Green Monday
            "2024-03-25",  # Greek Independence Day
            "2024-03-29",  # Catholic Good Friday
            "2024-04-01",  # Catholic Easter Monday / National Holiday
            "2024-05-01",  # Labour Day
            "2024-05-03",  # Orthodox Good Friday
            "2024-05-06",  # Orthodox Easter Monday
            "2024-05-07",  # Orthodox Easter Tuesday
            "2024-06-24",  # Holy Spirit Day
            "2024-08-15",  # Assumption Day
            "2024-10-01",  # Cyprus Independence Day
            "2024-10-28",  # Okhi Day
            "2024-12-24",  # Christmas Eve
            "2024-12-25",  # Christmas Day
            "2024-12-26",  # Boxing Day
        ]

    @pytest.fixture
    def non_holidays_sample(self):
        yield [
            "2024-01-02",  # Day after New Year's
            "2024-03-19",  # Day after Green Monday
            "2024-03-26",  # Day after Greek Independence Day
            "2024-05-02",  # Day after Labour Day
            "2024-05-08",  # Day after Orthodox Easter Tuesday
            "2024-06-18",  # Day after Holy Spirit Monday
            "2024-08-16",  # Day after Assumption Day
            "2024-10-02",  # Day after Cyprus Independence Day
            "2024-10-29",  # Day after Okhi Day
            "2024-12-27",  # Day after Boxing Day
        ]



import pytest

from exchange_calendars.exchange_calendar_xbog import XBOGExchangeCalendar
from .test_exchange_calendar import ExchangeCalendarTestBase


class TestXBOGCalendar(ExchangeCalendarTestBase):
    @pytest.fixture(scope="class")
    def calendar_cls(self):
        yield XBOGExchangeCalendar

    @pytest.fixture
    def max_session_hours(self):
        yield 6.5

    @pytest.fixture
    def regular_holidays_sample(self):
        yield [
            "2019-01-01",  # New Year's Day
            "2019-01-07",  # Epiphany
            "2019-03-25",  # St. Joseph's Day
            "2019-04-18",  # Maundy Thursday
            "2019-04-19",  # Good Friday
            "2019-05-01",  # Labour Day
            "2019-06-03",  # Ascension Day
            "2019-06-24",  # Corpus Christi
            "2019-07-01",  # Sacred Heart
            "2018-07-02",  # St. Peter and St. Paul Day
            "2018-07-20",  # Colombian Independence Day
            "2019-08-07",  # Battle of Boyaca
            "2019-08-19",  # Assumption Day
            "2019-10-14",  # Dia de la Raza
            "2019-11-04",  # All Saint's Day
            "2019-11-11",  # Cartagena Independence Day
            "2017-12-08",  # Immaculate Conception
            "2019-12-25",  # Christmas Day
            "2019-12-31",  # Last Trading Day
            #
            # Calendar has many holidays that are observed on the closest future
            # Monday. Ensure the future Monday is a holdiay.
            "2017-01-09",  # Epiphany moved to Monday, Jan 9
            "2019-03-25",  # St. Joseph's Day moved to Monday, Mar 25
            "2018-07-02",  # St. Peter and St. Paul Day moved to Monday, Jul 2
            "2018-08-20",  # Assumption Day moved to Monday, Aug 20
            "2018-10-15",  # Dia de la Raza moved to Monday, Oct 15
            "2018-11-05",  # All Saint's Day moved to Monday, Nov 5
            "2016-11-14",  # Cartagena Independence Day moved to Monday, Nov 14
            #
            # The last trading day of the year is a holiday for XBOG
            "2019-12-31",
            "2018-12-31",
            "2017-12-29",
            "2016-12-30",
        ]

    @pytest.fixture
    def non_holidays_sample(self):
        yield [
            # For holidays that fall on a weekend and are not made up, ensure
            # surrounding days are not treated as holidays.
            # New Years Day on Sunday, Jan 1st
            #   Note: 2011-12-30 is also a holiday (Last Trading Day).
            "2011-12-29",
            "2012-01-02",
            # Labour Day on Sunday, May 1st
            "2016-04-29",
            "2016-05-02",
            # Colombia Independence Day on Saturday, Jul 20
            "2019-07-19",
            "2019-07-22",
            # Battle of Boyaca on Sunday, Aug 7
            "2016-08-05",
            "2016-08-08",
            # Immaculate Conception on Sunday, Dec 8th.
            "2019-12-06",
            "2019-12-09",
            # Christmas on a Sunday
            "2016-12-23",
            "2016-12-26",
            #
            # Calendar has many holidays that are observed on the closest future
            # Monday. Ensure that, if falls on a non-Monday, the prior Monday
            # is not a holiday.
            "2017-01-06",  # Epiphany on Friday, Jan 6
            "2019-03-19",  # St. Joseph's Day on Tuesday, Mar 19
            "2018-06-29",  # St. Peter and St. Paul Day on Friday, Jun 29
            "2018-08-15",  # Assumption Day on Wednesday, Aug 15
            "2018-10-12",  # Dia de la Raza on Friday, Oct 12
            "2018-11-01",  # All Saint's Day on Thursday, Nov 1
            "2016-11-11",  # Cartagena Independence Day on Friday, Nov 11
        ]



import pytest

from exchange_calendars.exchange_calendar_xcse import XCSEExchangeCalendar
from .test_exchange_calendar import ExchangeCalendarTestBase


class TestXCSECalendar(ExchangeCalendarTestBase):
    @pytest.fixture(scope="class")
    def calendar_cls(self):
        yield XCSEExchangeCalendar

    @pytest.fixture
    def max_session_hours(self):
        # The XCSE is open from 9:00 am to 5:00 pm.
        yield 8

    @pytest.fixture
    def regular_holidays_sample(self):
        yield [
            # 2018
            "2018-01-01",  # New Year's Day
            "2018-03-29",  # Maundy Thursday
            "2018-03-30",  # Good Friday
            "2018-04-02",  # Easter Monday
            "2018-04-27",  # General Prayer Day
            "2018-05-10",  # Ascension Day
            "2018-05-11",  # Bank Holiday
            "2018-05-21",  # Whit Monday
            "2018-06-05",  # Constitution Day
            "2018-12-24",  # Christmas Eve
            "2018-12-25",  # Christmas Day
            "2018-12-26",  # Boxing Day
            "2018-12-31",  # New Year's Eve
            # First occurrence that regular holiday observed.
            "2009-05-22",  # May Bank Holiday
        ]

    @pytest.fixture
    def non_holidays_sample(self):
        yield [
            # Holidays that fall on a weekend are not made up. Ensure surrounding
            # days are not holidays.
            # In 2016, Constitution Day fell on a Sunday, so the market should
            # be open on both the prior Friday and the following Monday.
            "2016-06-03",
            "2016-06-06",
            # In 2010, Christmas fell on a Saturday, meaning Boxing Day fell on
            # a Sunday. The market should thus be open on the following Monday.
            "2010-12-27",
            # In 2017, New Year's Day fell on a Sunday, so the market should be
            # open on both the prior Friday and the following Monday.
            "2016-12-30",
            "2017-01-02",
            #
            # Year prior to first occurrence when May Bank Holiday observed.
            "2008-05-02",
        ]



from datetime import (
    time,
    datetime,
    timedelta,
)
from itertools import chain
from zoneinfo import ZoneInfo
import pandas as pd
from pandas.tseries.holiday import (
    Holiday,
    sunday_to_monday,
)
from .common_holidays import (
    new_years_day,
    orthodox_good_friday,
    orthodox_easter_monday,
)
from .exchange_calendar import HolidayCalendar, ExchangeCalendar


def sunday_to_tuesday(dt: datetime) -> datetime:
    """
    If holiday falls on Sunday, use Tuesday instead.
    """
    if dt.weekday() == 6:
        return dt + timedelta(2)
    return dt


NewYearsDay = new_years_day(observance=sunday_to_tuesday)
NewYearsDay2 = Holiday("New Year's Day", month=1, day=2, observance=sunday_to_monday)
OrthodoxChristmas = Holiday("Christmas Holiday", month=1, day=7)
StatehoodDay = Holiday(
    "Statehood Day of Serbia", month=2, day=15, observance=sunday_to_tuesday
)
StatehoodDay2 = Holiday(
    "Statehood Day of Serbia Holiday", month=2, day=16, observance=sunday_to_monday
)
OrthodoxGoodFriday = orthodox_good_friday()
OrthodoxEasterMonday = orthodox_easter_monday()
LabourDay = Holiday("Labour Day", month=5, day=1, observance=sunday_to_tuesday)
LabourDay2 = Holiday("Labour Day Holiday", month=5, day=2, observance=sunday_to_monday)
ArmisticeDay = Holiday("Armistice Day", month=11, day=11, observance=sunday_to_monday)


class XBELExchangeCalendar(ExchangeCalendar):
    """
    Calendar for the Belgrade Stock Exchange (Serbia).
    https://www.belex.rs/eng/trzista_i_hartije/kalendar

    Open Time: 9:30 AM
    Close Time: 2:00 PM

    Regularly-Observed Holidays:
      - New Year's Day (Jan 1)
      - Orthodox Christmas (Jan 7)
      - Statehood Day (Feb 15-16)
      - Orthodox Good Friday
      - Orthodox Easter Monday
      - Labour Day (May 1-2)
      - Armistice Day (Nov 11)
    """

    name = "XBEL"
    tz = ZoneInfo("Europe/Belgrade")
    open_times = ((None, time(9, 30)),)
    close_times = ((None, time(14, 00)),)

    @property
    def regular_holidays(self):
        return HolidayCalendar(
            [
                NewYearsDay,
                NewYearsDay2,
                OrthodoxChristmas,
                StatehoodDay,
                StatehoodDay2,
                OrthodoxGoodFriday,
                OrthodoxEasterMonday,
                LabourDay,
                LabourDay2,
                ArmisticeDay,
            ]
        )

    @property
    def adhoc_holidays(self):
        # Latest calendar: https://www.belex.rs/eng/trzista_i_hartije/kalendar
        # Historical calendars: https://www.belex.rs/eng/search_c/calendar
        # Exchange calendar 2025: https://www.belex.rs/data/2024/12/00140175_E.pdf
        # Exchange calendar 2024: https://www.belex.rs/data/2023/12/00135552_E.pdf
        # Exchange calendar 2023: https://www.belex.rs/data/2022/11/00130092_E.pdf
        # Exchange calendar 2022: https://www.belex.rs/data/2021/11/00124168_E.pdf
        # Exchange calendar 2021: https://www.belex.rs/data/2020/10/00117713_E.pdf
        misc_adhoc_holidays = [
            pd.Timestamp(
                "2025-12-31"
            ),  # Trading system maintenance, statistics and data migration
            pd.Timestamp(
                "2025-11-10"
            ),  # Relocation of CSDs servers and network equipment to a new address
            pd.Timestamp("2025-01-06"),  # Security improvement of the trading platform
            pd.Timestamp("2025-01-03"),  # Security improvement of the trading platform
            pd.Timestamp(
                "2024-12-31"
            ),  # Trading system maintenance, statistics and data migration
            pd.Timestamp(
                "2024-01-08"
            ),  # Day off for the second day of Orthodox Christmas
            pd.Timestamp(
                "2023-12-29"
            ),  # Trading system maintenance, statistics and data migration
            pd.Timestamp("2023-01-06"),  # Market closed
            pd.Timestamp("2023-01-05"),  # Market closed
            pd.Timestamp("2023-01-04"),  # Market closed
            pd.Timestamp(
                "2022-12-30"
            ),  # Trading system maintenance, statistics and data migration
            pd.Timestamp("2021-01-08"),  # Trading platform security upgrade
            pd.Timestamp(
                "2021-12-31"
            ),  # Trading system maintenance, statistics and data migration
        ]
        return list(
            chain(
                misc_adhoc_holidays,
            )
        )



# Copyright 2019 Quantopian, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import datetime
from collections.abc import Callable
from functools import partial
from itertools import chain
from zoneinfo import ZoneInfo
import pandas as pd
from pandas.tseries.holiday import (
    Holiday,
    nearest_workday,
    next_monday,
    previous_friday,
)

from .common_holidays import european_labour_day, new_years_day
from .exchange_calendar import (
    FRIDAY,
    MONDAY,
    THURSDAY,
    TUESDAY,
    ExchangeCalendar,
    HolidayCalendar,
)
from .lunisolar_holidays import (
    chinese_lunar_new_year_dates,
    dragon_boat_festival_dates,
    mid_autumn_festival_dates,
    qingming_festival_dates,
)

ONE_DAY = datetime.timedelta(1)


def check_after_2013(dt: datetime.datetime) -> bool:
    """Query if a date has a year component later than 2013."""
    return dt.year > 2013


def check_between_2013_2024(dt: datetime.datetime) -> bool:
    """Query if a date has a year component from 2014 thr 2023."""
    return 2024 > dt.year > 2013


def before_chinese_new_year_offset(holidays):
    """Offset holidays that fall on a weekend to the previous friday.

    Can be used to offset holidays that fall before Chinese New Year.
    """
    return pd.to_datetime(holidays.map(lambda d: previous_friday(d)))


def chinese_new_year_offset(holidays):
    """Offset holidays that fall on a weekend to the next monday.

    Can be used to offset holidays that fall after Chinese New Year.
    """
    return pd.to_datetime(holidays.map(lambda d: next_monday(d)))


def evalute_tomb_sweeping_weekday_extra(dt: pd.DatetimeIndex) -> pd.DatetimeIndex:
    """Evaluate extra tomb sweeping holdiays.

    Evalutes extra holidays when tomb sweeping day falls on a weekday.

    Reference: Taiwan Implementation Measures for Memorial Days and Holidays 1.5.2.1
    https://law.moj.gov.tw/LawClass/LawAll.aspx?pcode=D0020033
    """
    dts = []
    for d in dt:
        if d.year > 2012 and d.month == 4 and d.day == 4:
            if datetime.datetime(d.year, 4, 4).weekday() == THURSDAY:
                dts.append(datetime.datetime(d.year, 4, 5))
            else:
                dts.append(datetime.datetime(d.year, 4, 3))
    return pd.to_datetime(dts)


def nearest_weekday_from_2014(dt: datetime.datetime) -> datetime.datetime:
    """Return nearest weekday to a date that's later than 2013.

    If date is earlier than 2014 then returns `dt` as is.
    """
    return dt if dt.year < 2014 else nearest_workday(dt)


def holidays_nearest_weekday_from_2014(
    holidays: pd.DatetimeIndex,
) -> pd.DatetimeIndex:
    """Map `holidays` to nearest weekdays.

    holidays that are weekdays map to themselves.
    holidays before 2014 map to themselves.
    """
    return pd.to_datetime([nearest_weekday_from_2014(d) for d in holidays])


def nearest_weekday_2014_thr_2023(dt: datetime.datetime) -> datetime.datetime:
    """Return nearest weekday to a date from 2014 thr 2023.

    If date is of another year then returns `dt` as received.
    """
    return nearest_workday(dt) if 2024 > dt.year > 2013 else dt


def manual_nearest_weekday_2014_thr_2023(holidays):
    """Map holidays from 2013 thr 2023 to nearest weekday.

    holidays that are weekdays map to themselves.
    holidays earlier than 2014 or later than 2024 map to themselves.

    Can be used to implement nearest weekday observance rule for Chinese
    lunar calendar holidays from 2013 through 2023.
    """
    return pd.to_datetime(holidays.map(lambda d: nearest_weekday_2014_thr_2023(d)))


def manual_extra_days_07_thr_2023(holidays):
    """Return holidays to make up a four day weekend.

    Can be used to implement the four day weekend rule observed for Chinese
    lunar calendar holidays from 2007 through 2023.
    """
    friday_extras = [
        d + ONE_DAY
        for d in holidays
        if d.weekday() == THURSDAY and 2024 > d.year > 2006
    ]

    monday_extras = [
        d - ONE_DAY for d in holidays if d.weekday() == TUESDAY and 2024 > d.year > 2006
    ]

    return pd.to_datetime(friday_extras + monday_extras)


def bridge_mon(
    dt: datetime.datetime, checker: Callable = check_after_2013
) -> datetime.datetime | None:
    """Return previous Monday if date is a Tueday, otherwise None.

    Parameters
    ----------
    dt
        Date to query.

    checker
        Callable to implement additional checks. Should take a single
        parameter as a datetime.datetime and return a bool. `bridge_mon`
        will return None if `checker` returns False when passed `dt`.

        By default will check year of `dt` is later than 2013.
    """
    dt -= ONE_DAY
    return dt if (dt.weekday() == MONDAY and checker(dt)) else None


def bridge_fri(
    dt: datetime.datetime, checker: Callable = check_after_2013
) -> datetime.datetime | None:
    """Return following Friday if a date is a Thursday, otherwise None.

    Parameters
    ----------
    dt
        Date to query.

    checker
        Callable to implement additional checks. Should take a single
        parameter as a datetime.datetime and return a bool. `bridge_fri`
        will return None if `checker` returns False when passed `dt`.

        By default will check year of `dt` is later than 2013.
    """
    dt += ONE_DAY
    return dt if (dt.weekday() == FRIDAY and checker(dt)) else None


bridge_mon_2014_thr_2023 = partial(bridge_mon, checker=check_between_2013_2024)
bridge_fri_2014_thr_2023 = partial(bridge_fri, checker=check_between_2013_2024)

NewYearsDay = new_years_day(observance=nearest_weekday_from_2014)
NewYearsDayExtraMon = new_years_day(observance=bridge_mon)
NewYearsDayExtraFri = new_years_day(observance=bridge_fri_2014_thr_2023)

PeaceMemorialDay = Holiday(
    "Peace Memorial Day", month=2, day=28, observance=nearest_weekday_from_2014
)
PeaceMemorialDayExtraMon = Holiday(
    "Peace Memorial Day extra Monday",
    month=2,
    day=28,
    observance=bridge_mon_2014_thr_2023,
)
PeaceMemorialDayExtraFri = Holiday(
    "Peace Memorial Day extra Friday",
    month=2,
    day=28,
    observance=bridge_fri_2014_thr_2023,
)


WomenAndChildrensDay = Holiday(
    "Women and Children's Day",
    month=4,
    day=4,
    start_date="2011",
    observance=nearest_weekday_from_2014,
)
WomenAndChildrensDayExtraMon = Holiday(
    "Women and Children's Day extra Monday",
    month=4,
    day=4,
    start_date="2011",
    observance=bridge_mon_2014_thr_2023,
)
WomenAndChildrensDayExtraFri = Holiday(
    "Women and Children's Day extra Friday",
    month=4,
    day=4,
    start_date="2011",
    observance=bridge_fri_2014_thr_2023,
)


LabourDay = european_labour_day(observance=nearest_weekday_from_2014)

TeachersDay = Holiday(
    "Teachers' Day", month=9, day=28, start_date="2025", observance=nearest_workday
)

NationalDay = Holiday(
    "National Day of the Republic of China",
    month=10,
    day=10,
    observance=nearest_weekday_from_2014,
)
NationalDayExtraMon = Holiday(
    "National Day of the Republic of China extra Monday",
    month=10,
    day=10,
    observance=bridge_mon_2014_thr_2023,
)
NationalDayExtraFri = Holiday(
    "National Day of the Republic of China extra Friday",
    month=10,
    day=10,
    observance=bridge_fri_2014_thr_2023,
)

TaiwanRestorationDay = Holiday(
    "Taiwan Restoration Day",
    month=10,
    day=25,
    start_date="2025",
    observance=nearest_workday,
)

ConstitutionDay = Holiday(
    "Constitution Day",
    month=12,
    day=25,
    start_date="2025",
)

chinese_new_year = chinese_new_year_offset(chinese_lunar_new_year_dates)

chinese_new_years_eve = before_chinese_new_year_offset(
    chinese_new_year - ONE_DAY,
)

chinese_new_years_eve_2 = before_chinese_new_year_offset(
    chinese_new_years_eve - ONE_DAY,
)

chinese_new_year_2 = chinese_new_year_offset(
    chinese_new_year + ONE_DAY,
)

chinese_new_year_3 = chinese_new_year_offset(
    chinese_new_year_2 + ONE_DAY,
)

tomb_sweeping_day = manual_nearest_weekday_2014_thr_2023(qingming_festival_dates)
tomb_sweeping_day_weekday_extras = evalute_tomb_sweeping_weekday_extra(
    qingming_festival_dates
)
tomb_sweeping_day_weekend_extras = holidays_nearest_weekday_from_2014(
    qingming_festival_dates
)

dragon_boat_festival = manual_nearest_weekday_2014_thr_2023(dragon_boat_festival_dates)
dragon_boat_festival_weekday_extras = manual_extra_days_07_thr_2023(
    dragon_boat_festival
)
dragon_boat_festival_weekend_extras = holidays_nearest_weekday_from_2014(
    dragon_boat_festival_dates
)

mid_autumn_festival = manual_nearest_weekday_2014_thr_2023(
    mid_autumn_festival_dates,
)
mid_autumn_festival_weekday_extras = manual_extra_days_07_thr_2023(mid_autumn_festival)
mid_autumn_festival_weekend_extras = holidays_nearest_weekday_from_2014(
    mid_autumn_festival_dates
)

# Taiwan takes multiple days off before and after chinese new year,
# and sometimes it is unclear precisely which days will be holidays.
chinese_new_year_extras = pd.to_datetime(
    [
        "2002-02-07",
        "2002-02-15",
        "2003-01-29",
        "2004-01-19",
        "2005-02-04",
        "2006-02-02",
        "2007-02-22",
        "2007-02-23",
        "2008-02-04",
        "2009-01-29",
        "2009-01-30",
        "2010-02-18",
        "2010-02-19",
        "2011-01-31",
        "2012-01-26",
        "2012-01-27",
        "2013-02-14",
        "2013-02-15",
        "2014-01-28",
        "2015-02-16",
        "2016-02-11",
        "2016-02-12",
        "2017-01-25",
        "2018-02-13",
        "2019-01-31",
        "2019-02-08",
        "2020-01-21",
        "2020-01-22",
        "2021-02-08",
        "2021-02-09",
        "2022-01-27",
        "2023-01-26",
        "2023-01-27",
        "2024-02-06",
        "2024-02-07",
        "2025-01-23",
        "2025-01-24",
        "2026-02-20",
    ]
)

# Some abnormal observances of regularly observed holidays.
extra_holidays = pd.to_datetime(
    [
        "2026-02-12",  # No trading
        "2021-04-02",  # Women And Childrens Day
        "2020-04-02",  # Tomb Sweeping Day
        "2016-04-05",  # Tomb Sweeping Day
        "2012-12-31",  # New Year's Eve
        "2012-02-27",  # Peace Memorial Day
        "2009-01-02",  # New Year's Day
        "2006-10-09",  # National Day
        "2005-09-01",  # Bank Holiday
        "2018-04-06",  # Tomb Sweeping Day, https://www.dgpa.gov.tw/information?uid=41&pid=7488
        "2007-04-06",  # Tomb Sweeping Day
    ]
)

typhoons = pd.to_datetime(
    [
        "2024-10-03",
        "2024-10-02",
        "2024-07-25",
        "2024-07-24",
        "2023-08-03",
        "2019-09-30",
        "2019-08-09",
        "2016-09-28",
        "2016-09-27",
        "2016-07-08",
        "2015-09-29",
        "2015-07-10",
        "2014-07-23",
        "2013-08-21",
        "2012-08-02",
        "2009-08-07",
        "2008-09-29",
        "2008-07-28",
        "2007-09-18",
        "2005-08-05",
        "2005-07-18",
        "2004-10-25",
        "2004-08-25",
        "2004-08-24",
        "2002-09-06",
    ]
)


class XTAIExchangeCalendar(ExchangeCalendar):
    """
    Exchange calendar for the Taiwan Stock Exchange Corporation (XTAI).
    https://www.twse.com.tw/en/trading/holiday.html

    Open Time: 9:00 AM, CST
    Close Time: 1:30 PM, CST

    Regularly-Observed Holidays:
    - New Year's Day
    - Chinese New Year's Eve
    - Chinese New Year
    - Peace Memorial Day (Feb 28)
    - Women and Children's Day (Apr 4)
    - Tomb Sweeping Day (Lunar Calendar)
    - Labour Day (May 1)
    - Dragon Boat Festival (Lunar Calendar)
    - Mid-Autumn Festival (Lunar Calendar)
    - National Day of the Republic of China (Oct 10)


    Early Closes:
    - None
    """

    name = "XTAI"

    tz = ZoneInfo("Asia/Taipei")

    open_times = ((None, datetime.time(9)),)

    close_times = ((None, datetime.time(13, 30)),)

    @property
    def regular_holidays(self):
        return HolidayCalendar(
            [
                NewYearsDay,
                NewYearsDayExtraMon,
                NewYearsDayExtraFri,
                PeaceMemorialDay,
                PeaceMemorialDayExtraMon,
                PeaceMemorialDayExtraFri,
                WomenAndChildrensDay,
                WomenAndChildrensDayExtraMon,
                WomenAndChildrensDayExtraFri,
                LabourDay,
                TeachersDay,
                NationalDay,
                NationalDayExtraMon,
                NationalDayExtraFri,
                TaiwanRestorationDay,
                ConstitutionDay,
            ]
        )

    @property
    def adhoc_holidays(self):
        return list(
            chain(
                extra_holidays,
                typhoons,
                chinese_new_years_eve,
                chinese_new_years_eve_2,
                chinese_new_year,
                chinese_new_year_2,
                chinese_new_year_3,
                chinese_new_year_extras,
                tomb_sweeping_day,
                tomb_sweeping_day_weekday_extras,
                tomb_sweeping_day_weekend_extras,
                dragon_boat_festival,
                dragon_boat_festival_weekday_extras,
                dragon_boat_festival_weekend_extras,
                mid_autumn_festival,
                mid_autumn_festival_weekday_extras,
                mid_autumn_festival_weekend_extras,
            )
        )



#
# Copyright 2018 Quantopian, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Useful source for the history of New Zealand public holidays:
# Nancy Swarbrick, 'Public holidays', Te Ara - the Encyclopedia of New Zealand
# http://www.TeAra.govt.nz/en/public-holidays/print (accessed 22 December 2019)

from datetime import time
from zoneinfo import ZoneInfo

from pandas import Timestamp
from pandas.tseries.holiday import (
    MO,
    DateOffset,
    EasterMonday,
    GoodFriday,
    Holiday,
    next_monday,
    next_monday_or_tuesday,
    previous_workday,
    weekend_to_monday,
)

from .common_holidays import anzac_day, boxing_day, christmas, new_years_day
from .exchange_calendar import HolidayCalendar, ExchangeCalendar

# Prior to 2015, Waitangi Day and Anzac Day are not "Mondayized",
# that is, if they occur on the weekend, there is no make-up.
MONDAYIZATION_START_DATE = "2015-01-01"

# Regular Holidays
# ----------------
NewYearsDay = new_years_day(observance=next_monday)

DayAfterNewYearsDay = Holiday(
    "Day after New Year's Day",
    month=1,
    day=2,
    observance=next_monday_or_tuesday,
)

WaitangiDayNonMondayized = Holiday(
    "Waitangi Day",
    month=2,
    day=6,
    end_date=MONDAYIZATION_START_DATE,
)

WaitangiDay = Holiday(
    "Waitangi Day",
    month=2,
    day=6,
    observance=weekend_to_monday,
    start_date=MONDAYIZATION_START_DATE,
)

AnzacDayNonMondayized = anzac_day(end_date=MONDAYIZATION_START_DATE)

AnzacDay = anzac_day(
    observance=weekend_to_monday,
    start_date=MONDAYIZATION_START_DATE,
)

QueensBirthday = Holiday(
    "Queen's Birthday",
    month=6,
    day=1,
    offset=DateOffset(weekday=MO(1)),
)

LabourDay = Holiday(
    "Labour Day",
    month=10,
    day=1,
    offset=DateOffset(weekday=MO(4)),
)

Christmas = christmas(observance=next_monday)

BoxingDay = boxing_day(observance=next_monday_or_tuesday)


# Early Closes
# ------------
BusinessDayPriorToChristmasDay = Holiday(
    "Business Day prior to Christmas Day",
    month=12,
    day=25,
    observance=previous_workday,
    start_date="2011-01-01",
)

BusinessDayPriorToNewYearsDay = Holiday(
    "Business Day prior to New Year's Day",
    month=1,
    day=1,
    observance=previous_workday,
    start_date="2011-01-01",
)

# Further ad-hoc holidays and closures sourced from gaps in index series
XMAS22DEC86 = Timestamp("1986-12-22")
XMAS23DEC86 = Timestamp("1986-12-23")
XMAS24DEC86 = Timestamp("1986-12-24")
XMAS29DEC86 = Timestamp("1986-12-29")
XMAS30DEC86 = Timestamp("1986-12-30")
XMAS31DEC86 = Timestamp("1986-12-31")
XMAS24DEC90 = Timestamp("1990-12-24")
XMAS31DEC90 = Timestamp("1990-12-31")
XMAS27DEC91 = Timestamp("1991-12-27")
NYDX03JAN92 = Timestamp("1992-01-03")
XMAS31DEC99 = Timestamp("1999-12-31")
QueenMourning2022 = Timestamp("2022-09-26")

# Matariki Day. Is set by Maori Lunar system which fluctuates relative to
# Gregorian so no set date each year.
# https://www.mbie.govt.nz/business-and-employment/employment-and-skills/employment-legislation-reviews/matariki/matariki-public-holiday/
MatarikiDayDates = {}
MatarikiDayDates["2022"] = Timestamp("2022-06-24")
MatarikiDayDates["2023"] = Timestamp("2023-07-14")
MatarikiDayDates["2024"] = Timestamp("2024-06-28")
MatarikiDayDates["2025"] = Timestamp("2025-06-20")
MatarikiDayDates["2026"] = Timestamp("2026-07-10")
MatarikiDayDates["2027"] = Timestamp("2027-06-25")
MatarikiDayDates["2028"] = Timestamp("2028-07-14")
MatarikiDayDates["2029"] = Timestamp("2029-07-06")
MatarikiDayDates["2030"] = Timestamp("2030-06-21")
MatarikiDayDates["2031"] = Timestamp("2031-07-11")
MatarikiDayDates["2032"] = Timestamp("2032-07-02")
MatarikiDayDates["2033"] = Timestamp("2033-06-24")
MatarikiDayDates["2034"] = Timestamp("2034-07-07")
MatarikiDayDates["2035"] = Timestamp("2035-06-29")
MatarikiDayDates["2036"] = Timestamp("2036-07-21")
MatarikiDayDates["2037"] = Timestamp("2037-07-10")
MatarikiDayDates["2038"] = Timestamp("2038-06-25")
MatarikiDayDates["2039"] = Timestamp("2039-07-15")
MatarikiDayDates["2040"] = Timestamp("2040-07-06")
MatarikiDayDates["2041"] = Timestamp("2041-07-19")
MatarikiDayDates["2042"] = Timestamp("2042-07-11")
MatarikiDayDates["2043"] = Timestamp("2043-07-03")
MatarikiDayDates["2044"] = Timestamp("2044-06-24")
MatarikiDayDates["2045"] = Timestamp("2045-07-07")
MatarikiDayDates["2046"] = Timestamp("2046-06-29")
MatarikiDayDates["2047"] = Timestamp("2047-07-19")
MatarikiDayDates["2048"] = Timestamp("2048-07-03")
MatarikiDayDates["2049"] = Timestamp("2049-06-25")


class XNZEExchangeCalendar(ExchangeCalendar):
    """
    Exchange calendar for the New Zealand Exchange (NZX).

    Open Time: 10:00 AM, NZ
    Close Time: 4:45 PM, NZ

    Regularly-Observed Holidays:
    - New Year's Day
    - Day after New Year's Day
    - Waitangi Day
    - Good Friday
    - Easter Monday
    - Anzac Day
    - Queen's Birthday
    - Labour Day
    - Christmas
    - Boxing Day

    Early Closes:
    - Business Day prior to Christmas Day
    - Business Day prior to New Year's Day
    """

    regular_early_close = time(12, 45)

    name = "XNZE"

    tz = ZoneInfo("Pacific/Auckland")

    open_times = ((None, time(10)),)

    close_times = ((None, time(16, 45)),)

    @property
    def regular_holidays(self):
        return HolidayCalendar(
            [
                NewYearsDay,
                DayAfterNewYearsDay,
                WaitangiDayNonMondayized,
                WaitangiDay,
                GoodFriday,
                EasterMonday,
                AnzacDayNonMondayized,
                AnzacDay,
                QueensBirthday,
                LabourDay,
                Christmas,
                BoxingDay,
            ]
        )

    # add extra ad-hoc holidays
    @property
    def adhoc_holidays(self):
        dates = [
            XMAS22DEC86,
            XMAS23DEC86,
            XMAS24DEC86,
            XMAS29DEC86,
            XMAS30DEC86,
            XMAS31DEC86,
            XMAS24DEC90,
            XMAS31DEC90,
            XMAS27DEC91,
            NYDX03JAN92,
            XMAS31DEC99,
            QueenMourning2022,
        ]
        dates += MatarikiDayDates.values()
        return dates

    @property
    def special_closes(self):
        return [
            (
                self.regular_early_close,
                HolidayCalendar(
                    [BusinessDayPriorToChristmasDay, BusinessDayPriorToNewYearsDay]
                ),
            )
        ]
